#!/usr/bin/python
import os
import sys
import copy
import logging
import subprocess
import argparse
import pprint
import shutil
from yggdrasil.constants import LANGUAGES, LANGUAGES_WITH_ALIASES


logger = logging.getLogger(__name__)
package_dir = os.path.dirname(os.path.abspath(__file__))


def githook():
    r"""Git hook to determine if the Github workflow need to be
    re-generated."""
    try:
        files = subprocess.check_output(
            ["git", "diff-index", "--cached", "--name-only",
             "--diff-filter=ACMRTUXB", "HEAD"],
            stderr=subprocess.PIPE).decode('utf-8').splitlines()
    except subprocess.CalledProcessError:
        return 1
    regen = (os.path.join('utils', 'test-install-base.yml') in files)
    if regen:
        try:
            gitdir = subprocess.check_output(
                ["git", "rev-parse", "--git-dir"],
                stderr=subprocess.PIPE).decode('utf-8').strip()
        except subprocess.CalledProcessError:
            return 1
        workflow_dir = os.path.join(gitdir, '..', '.github', 'workflows')
        generate_gha_workflow(args=[], gitdir=gitdir)
        try:
            subprocess.run(
                ['git', 'add',
                 os.path.join(workflow_dir, 'test-install.yml')],
                check=True,
                stderr=subprocess.STDOUT,
                stdout=subprocess.PIPE)
        except subprocess.CalledProcessError:
            return 1
    return 0


class ArgumentTuple(tuple):

    def __new__(self, args, kwargs):
        return tuple.__new__(ArgumentTuple, (args, kwargs))


class ConditionalArgumentTuple(ArgumentTuple):

    def __new__(cls, args, kwargs, conditions=None):
        out = ArgumentTuple.__new__(ConditionalArgumentTuple,
                                    args, kwargs)
        out.conditions = conditions
        if out.conditions is None:
            out.conditions = {}
        return out
            

class ArgumentBase(object):
    
    def __init__(self, arguments=None, conditions=None, **kwargs):
        self.arguments = arguments
        if self.arguments is None:
            self.arguments = []
        self.conditions = conditions
        if self.conditions is None:
            self.conditions = {}
        self.kwargs = kwargs


class ArgumentParser(ArgumentBase):
    pass


class ArgumentGroup(ArgumentBase):

    def __init__(self, exclusive=False, **kwargs):
        self.exclusive = exclusive
        super(ArgumentGroup, self).__init__(**kwargs)


class ArgumentSubparser(ArgumentBase):

    def __init__(self, parsers=None, **kwargs):
        self.parsers = parsers
        if self.parsers is None:
            self.parsers = []
        super(ArgumentSubparser, self).__init__(**kwargs)


class SubCommandMeta(type):
    r"""Meta class for subcommands."""

    def __call__(cls, *args, **kwargs):
        return cls.call(*args, **kwargs)


def ReplacementWarning(old, new):
    import warnings
    warnings.warn(("'%s' will soon be removed. Use '%s' instead.")
                  % (old, new), FutureWarning)


class SubCommand(metaclass=SubCommandMeta):
    r"""Class for handling subcommands so that they can be run
    as subcommands or individually."""

    name = None
    help = None
    arguments = []
    allow_unknown = False

    @classmethod
    def parse_args(cls, parser, args=None, allow_unknown=False):
        # TODO: Check choices for positional arguments that can
        # have more than one element
        if isinstance(args, argparse.Namespace):
            return args
        if cls.allow_unknown or allow_unknown:
            args, extra = parser.parse_known_args(args=args)
            args._extra_commands = extra
        else:
            args = parser.parse_args(args=args)
        for k in ['language', 'languages']:
            v = getattr(args, k, None)
            if isinstance(v, list):
                v_flag = getattr(args, k + '_flag', None)
                if isinstance(v_flag, list):
                    v.extend(v_flag)
                if (len(v) == 0) or ('all' in v):
                    setattr(args, k, LANGUAGES['all'])
                    args.all_languages = True
        return args

    @classmethod
    def func(cls, args):  # pragma: debug
        raise NotImplementedError

    @classmethod
    def call(cls, args=None, **kwargs):
        parser = cls.get_parser(args=args)
        args = cls.parse_args(parser, args=args)
        return cls.func(args, **kwargs)

    @classmethod
    def get_parser(cls, args=None):
        parser = argparse.ArgumentParser(cls.help)
        cls.add_arguments(parser, args=args)
        return parser

    @classmethod
    def add_argument_to_parser(cls, parser, x):
        if hasattr(x, 'conditions') and x.conditions:
            for k, v in x.conditions.items():
                if k == 'os':
                    from yggdrasil import platform
                    if platform._platform not in v:
                        return
                else:  # pragma: debug
                    raise NotImplementedError(k)
        if isinstance(x, list):
            for xx in x:
                cls.add_argument_to_parser(parser, xx)
        elif isinstance(x, (tuple, ArgumentTuple,
                            ConditionalArgumentTuple)):
            assert(len(x) == 2)
            args, kwargs = x[:]
            try:
                parser.add_argument(*args, **kwargs)
            except ValueError:
                if kwargs.get('action', None) == 'extend':
                    kwargs['action'] = 'append'
                    kwargs.pop('nargs', None)
                    parser.add_argument(*args, **kwargs)
                else:
                    raise
        elif isinstance(x, ArgumentGroup):
            if x.exclusive:
                group = parser.add_mutually_exclusive_group(**x.kwargs)
            else:
                group = parser.add_argument_group(**x.kwargs)
            cls.add_argument_to_parser(group, x.arguments)
        elif isinstance(x, ArgumentSubparser):
            subparsers = parser.add_subparsers(**x.kwargs)
            for xx in x.parsers:
                isubparser = subparsers.add_parser(**xx.kwargs)
                cls.add_argument_to_parser(isubparser, x.arguments)
                cls.add_argument_to_parser(isubparser, xx.arguments)
        else:
            raise NotImplementedError("type(x) = %s" % type(x))

    @classmethod
    def add_arguments(cls, parser, args=None):
        cls.add_argument_to_parser(parser, cls.arguments)

    @classmethod
    def add_subparser(cls, subparsers, args=None):
        parser = subparsers.add_parser(cls.name, help=cls.help)
        cls.add_arguments(parser, args=args)
        parser.set_defaults(func=cls.func)


class main(SubCommand):
    r"""Runner for yggdrasil CLI."""

    name = "yggdrasil"
    help = (
        "Command line interface for the yggdrasil package.")
    arguments = []

    @classmethod
    def get_parser(cls, **kwargs):
        from yggdrasil import __version__ as ver
        parser = super(main, cls).get_parser(**kwargs)
        parser.add_argument('--version', action='version',
                            version=('yggdrasil %s' % ver))
        subparsers = parser.add_subparsers(title='subcommands',
                                           dest='subcommand')
        parser._ygg_subparsers = {}
        for x in [yggrun, ygginfo, validate_yaml,
                  yggcc, yggcompile, yggclean,
                  ygginstall, update_config,
                  regen_metaschema, regen_schema,
                  yggmodelform, yggdevup, run_tsts,
                  timing_plots, generate_gha_workflow]:
            x.add_subparser(subparsers, args=kwargs.get('args', None))
            parser._ygg_subparsers[x.name] = x
        return parser

    @classmethod
    def parse_args(cls, parser, args=None, **kwargs):
        if args is None:
            args = sys.argv[1:]
        if isinstance(args, list) and ('test' in args):
            kwargs['allow_unknown'] = True
        args = super(main, cls).parse_args(parser, args=args, **kwargs)
        if args.subcommand:
            args = parser._ygg_subparsers[args.subcommand].parse_args(
                parser, args=args, **kwargs)
        return args

    @classmethod
    def func(cls, args):
        args.func(args)


class yggrun(SubCommand):
    r"""Start a run."""

    name = "run"
    help = "Run an integration."
    arguments = [
        (('yamlfile', ),
         {'nargs': '+',
          'help': "One or more yaml specification files."})]

    @classmethod
    def add_arguments(cls, parser, **kwargs):
        from yggdrasil import config
        super(yggrun, cls).add_arguments(parser, **kwargs)
        config.get_config_parser(parser, skip_sections='testing')

    @classmethod
    def func(cls, args):
        from yggdrasil import runner, config
        prog = sys.argv[0].split(os.path.sep)[-1]
        with config.parser_config(args):
            runner.run(args.yamlfile, ygg_debug_prefix=prog,
                       production_run=args.production_run)


class ygginfo(SubCommand):
    r"""Print information about yggdrasil installation."""

    name = 'info'
    help = ('Display information about the current yggdrasil '
            'installation.')
    arguments = [
        (('--no-languages', ),
         {'action': 'store_true', 'dest': 'no_languages',
          'help': ('Don\'t print information about individual '
                   'languages.')}),
        (('--no-comms', ),
         {'action': 'store_true', 'dest': 'no_comms',
          'help': ('Don\'t print information about individual '
                   'comms.')}),
        (('--verbose', '-v'),
         {'action': 'store_true',
          'help': ('Increase the verbosity of the printed '
                   'information.')}),
        ArgumentSubparser(
            title='tool', dest='tool',
            description='Compilation tool types to get info about.',
            arguments=[
                (('language', ),
                 {'choices': LANGUAGES_WITH_ALIASES['compiled'],
                  'type': str.lower,
                  'help': 'Language to get tool information for.'}),
                (('--toolname', ),
                 {'default': None,
                  'help': ('Name of tool to get information for. '
                           'If not provided, information for the '
                           'default tool will be returned.')}),
                (('--flags', ),
                 {'action': 'store_true',
                  'help': ('Display the flags that yggdrasil will '
                           ' pass to the tool when it is called.')}),
                (('--fullpath', ),
                 {'action': 'store_true',
                  'help': 'Get the full path to the tool exectuable.'})],
            parsers=[
                ArgumentParser(
                    name='compiler',
                    help='Get information about a compiler.'),
                ArgumentParser(
                    name='linker',
                    help='Get information about a linker.',
                    arguments=[
                        (('--library', ),
                         {'action': 'store_true',
                          'help': 'Get flags for linking a library.'})]),
                ArgumentParser(
                    name='archiver',
                    help='Get information about a archiver.')])]

    @classmethod
    def func(cls, args, return_str=False):
        from yggdrasil import platform
        from yggdrasil.components import import_component
        if args.tool:
            drv = import_component('model', args.language)
            if args.flags:
                if args.tool == 'compiler':
                    flags = drv.get_compiler_flags(
                        for_model=True, toolname=args.toolname,
                        dry_run=True, dont_link=True)
                    if '/link' in flags:  # pragma: windows
                        flags = flags[:flags.index('/link')]
                    for k in ['-c']:
                        if k in flags:
                            flags.remove(k)
                else:
                    if args.tool == 'archiver':
                        libtype = 'static'
                    elif getattr(args, 'library', False):
                        libtype = 'shared'
                    else:
                        libtype = 'object'
                    flags = drv.get_linker_flags(
                        for_model=True, toolname=args.toolname,
                        dry_run=True, libtype=libtype)
                out = ' '.join(flags)
                if platform._is_win:  # pragma: windows:
                    out = out.replace('/', '-')
                    out = out.replace('\\', '/')
            elif args.fullpath:
                out = drv.get_tool(args.tool).get_executable(full_path=True)
            else:
                out = drv.get_tool(args.tool, return_prop='name')
            if return_str:
                return out
            print(out)
            return
        from yggdrasil import tools, config, __version__
        lang_list = tools.get_installed_lang()
        comm_list = tools.get_installed_comm()
        comm_lang_list = []
        prefix = '    '
        curr_prefix = ''
        vardict = [
            ('Location', os.path.dirname(__file__)),
            ('Version', __version__),
            ('Languages', ', '.join(lang_list)),
            ('Communication Mechanisms',
             ', '.join(tools.get_installed_comm())),
            ('Default Comm Mechanism', tools.get_default_comm()),
            ('Config File', config.usr_config_file)]
        try:
            # Add language information
            if not args.no_languages:
                # Install languages
                vardict.append(('Installed Languages:', ''))
                curr_prefix += prefix
                for lang in sorted(lang_list):
                    drv = import_component('model', lang)
                    vardict.append(
                        (curr_prefix + '%s:' % lang.upper(), ''))
                    if not drv.comms_implicit:
                        comm_lang_list.append(lang)
                    curr_prefix += prefix
                    exec_name = drv.language_executable()
                    if exec_name:
                        if not os.path.isabs(exec_name):
                            exec_name = shutil.which(exec_name)
                        vardict.append((curr_prefix + 'Location',
                                        exec_name))
                    vardict.append((curr_prefix + 'Version',
                                    drv.language_version()))
                    curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                # Not installed languages
                vardict.append(("Languages Not Installed:", ''))
                curr_prefix += prefix
                for lang in tools.get_supported_lang():
                    if lang in lang_list:
                        continue
                    drv = import_component('model', lang)
                    vardict.append(
                        (curr_prefix + '%s:' % lang.upper(), ''))
                    curr_prefix += prefix
                    vardict.append(
                        (curr_prefix + "Language Installed",
                         drv.is_language_installed()))
                    if drv.executable_type == 'compiler':
                        curr_prefix += prefix
                        vardict += [
                            (curr_prefix
                             + ("%s Installed (%s)"
                                % (x.title(),
                                   getattr(drv, 'default_%s' % x, None))),
                             drv.is_tool_installed(x))
                            for x in ['compiler', 'linker', 'archiver']]
                        curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                    vardict.append(
                        (curr_prefix + "Base Languages Installed",
                         drv.are_base_languages_installed()))
                    missing = []
                    if not drv.are_base_languages_installed(
                            missing=missing):
                        vardict.append(
                            (curr_prefix
                             + "Base Languages Not Installed",
                             missing))
                    vardict.append(
                        (curr_prefix + "Dependencies Installed",
                         drv.are_dependencies_installed()))
                    if not drv.are_dependencies_installed():
                        vardict.append(
                            (curr_prefix
                             + "Dependencies Not Installed",
                             [b for b in drv.interface_dependencies if
                              (not drv.is_library_installed(b))]))
                    vardict.append(
                        (curr_prefix + "Interface Installed",
                         drv.is_interface_installed()))
                    vardict.append((curr_prefix + "Comm Installed",
                                    drv.is_comm_installed()))
                    vardict.append((curr_prefix + "Configured",
                                    drv.is_configured()))
                    vardict.append((curr_prefix + "Disabled",
                                    drv.is_disabled()))
                    curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
            # Add comm information
            if not args.no_comms:
                # Fully installed comms
                vardict.append(
                    ('Comms Available for All Languages:', ''))
                curr_prefix += prefix
                for comm in sorted(comm_list):
                    cmm = import_component('comm', comm)
                    vardict.append(
                        (curr_prefix + '%s' % comm.upper(), ''))
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                # Partially installed comms
                vardict.append(
                    ('Comms Available for Some/No Languages:', ''))
                curr_prefix += prefix
                for comm in tools.get_supported_comm():
                    if comm in comm_list:
                        continue
                    cmm = import_component('comm', comm)
                    vardict.append(
                        (curr_prefix + '%s:' % comm.upper(), ''))
                    curr_prefix += prefix
                    avail = [cmm.is_installed(language=lang)
                             for lang in comm_lang_list]
                    vardict.append(
                        (curr_prefix + "Available for ",
                         sorted([comm_lang_list[i].upper()
                                 for i in range(len(avail))
                                 if avail[i]])))
                    vardict.append(
                        (curr_prefix + "Not Available for ",
                         sorted([comm_lang_list[i].upper()
                                 for i in range(len(avail))
                                 if not avail[i]])))
                    curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
            # Add verbose information
            if args.verbose:
                # Path variables
                path_vars = ['PATH', 'C_INCLUDE_PATH', 'INCLUDE',
                             'LIBRARY_PATH', 'LD_LIBRARY_PATH', 'LIB']
                vardict.append(('Environment paths:', ''))
                curr_prefix += prefix
                for k in path_vars:
                    if os.environ.get(k, ''):
                        vardict.append(
                            (curr_prefix + k, '\n%s%s'
                             % (curr_prefix + prefix,
                                ("\n" + curr_prefix + prefix).join(
                                    os.environ[k].split(os.pathsep)))))
                    else:
                        vardict.append((curr_prefix + k, ''))
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                # Environment variabless
                env_vars = ['CONDA_PREFIX', 'CONDA', 'SDKROOT', 'CC',
                            'CXX', 'FC', 'GFORTRAN', 'DISPLAY', 'CL',
                            '_CL_']
                if platform._is_win:  # pragma: windows
                    env_vars += ['VCPKG_ROOT']
                vardict.append(('Environment variables:', ''))
                curr_prefix += prefix
                for k in env_vars:
                    vardict.append(
                        (curr_prefix + k, os.environ.get(k, None)))
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                # Conda info
                if os.environ.get('CONDA_PREFIX', ''):
                    if platform._is_win:  # pragma: windows
                        out = tools.bytes2str(subprocess.check_output(
                            'conda info', shell=True)).strip()
                    else:
                        out = tools.bytes2str(subprocess.check_output(
                            ['conda', 'info'])).strip()
                    curr_prefix += prefix
                    vardict.append(
                        (curr_prefix + 'Conda Info:', "\n%s%s"
                         % (curr_prefix + prefix,
                            ("\n" + curr_prefix + prefix).join(
                                out.splitlines(False)))))
                    curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                # Configuration
                with open(config.usr_config_file, 'r') as fd:
                    contents = fd.read()
                vardict.append(
                    ('Configuration file:', '%s\n\t%s' % (
                        config.usr_config_file,
                        '\n\t'.join(contents.splitlines()))))
                # R and reticulate info
                Rdrv = import_component("model", "R")
                if Rdrv.is_installed():
                    env_reticulate = copy.deepcopy(os.environ)
                    env_reticulate['RETICULATE_PYTHON'] = sys.executable
                    # Stack size
                    out = Rdrv.run_executable(
                        ["-e", "Cstack_info()"]).strip()
                    vardict.append(
                        (curr_prefix + "R Cstack_info:", "\n%s%s"
                         % (curr_prefix + prefix,
                            ("\n" + curr_prefix + prefix).join(
                                out.splitlines(False)))))
                    # Compilation tools
                    interp = 'R'.join(
                        Rdrv.get_interpreter().rsplit('Rscript', 1))
                    vardict.append(
                        (curr_prefix + "R C Compiler:", ""))
                    curr_prefix += prefix
                    for x in ['CC', 'CFLAGS', 'CXX', 'CXXFLAGS']:
                        try:
                            out = tools.bytes2str(
                                subprocess.check_output(
                                    [interp, 'CMD', 'config', x],
                                    stderr=subprocess.STDOUT)).strip()
                        except subprocess.CalledProcessError:
                            out = 'ERROR (missing Rtools?)'
                        vardict.append(
                            (curr_prefix + x, "%s"
                             % ("\n" + curr_prefix + prefix).join(
                                 out.splitlines(False))))
                    curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                    # Session info
                    out = Rdrv.run_executable(
                        ["-e", "sessionInfo()"]).strip()
                    vardict.append(
                        (curr_prefix + "R sessionInfo:", "\n%s%s"
                         % (curr_prefix + prefix,
                            ("\n" + curr_prefix + prefix).join(
                                out.splitlines(False)))))
                    # Reticulate conda_list
                    if os.environ.get('CONDA_PREFIX', ''):
                        if platform._is_win:  # pragma: windows
                            out = tools.bytes2str(subprocess.check_output(
                                'conda info --json', shell=True)).strip()
                        else:
                            out = tools.bytes2str(subprocess.check_output(
                                ['conda', 'info', '--json'])).strip()
                        vardict.append(
                            (curr_prefix
                             + "conda info --json",
                             "\n%s%s" % (
                                 curr_prefix + prefix,
                                 ("\n" + curr_prefix + prefix).join(
                                     out.splitlines(False)))))
                        try:
                            out = Rdrv.run_executable(
                                ["-e", ("library(reticulate); "
                                        "reticulate::conda_list()")],
                                env=env_reticulate).strip()
                        except BaseException:  # pragma: debug
                            out = 'ERROR'
                        vardict.append(
                            (curr_prefix
                             + "R reticulate::conda_list():",
                             "\n%s%s" % (
                                 curr_prefix + prefix,
                                 ("\n" + curr_prefix + prefix).join(
                                     out.splitlines(False)))))
                    # Windows python versions
                    if platform._is_win:  # pragma: windows
                        out = Rdrv.run_executable(
                            ["-e",
                             ("library(reticulate); "
                              "reticulate::py_versions_windows()")],
                            env=env_reticulate).strip()
                        vardict.append(
                            (curr_prefix
                             + "R reticulate::py_versions_windows():",
                             "\n%s%s" % (
                                 curr_prefix + prefix,
                                 ("\n" + curr_prefix + prefix).join(
                                     out.splitlines(False)))))
                    # conda_binary
                    if platform._is_win and shutil.which('conda'):  # pragma: windows
                        out = Rdrv.run_executable(
                            ["-e",
                             ("library(reticulate); "
                              "conda <- reticulate:::conda_binary(\"auto\"); "
                              "system(paste(conda, \"info --json\"))")],
                            env=env_reticulate).strip()
                        vardict.append(
                            (curr_prefix
                             + "R reticulate::py_versions_windows():",
                             "\n%s%s" % (
                                 curr_prefix + prefix,
                                 ("\n" + curr_prefix + prefix).join(
                                     out.splitlines(False)))))
                    # Reticulate py_config
                    out = Rdrv.run_executable(
                        ["-e", ("library(reticulate); "
                                "reticulate::py_config()")],
                        env=env_reticulate).strip()
                    vardict.append(
                        (curr_prefix + "R reticulate::py_config():",
                         "\n%s%s" % (
                             curr_prefix + prefix,
                             ("\n" + curr_prefix + prefix).join(
                                 out.splitlines(False)))))
        finally:
            # Print things
            max_len = max(len(x[0]) for x in vardict)
            lines = []
            line_format = '%-' + str(max_len) + 's' + prefix + '%s'
            for k, v in vardict:
                lines.append(line_format % (k, v))
            msg = "yggdrasil info:\n%s" % '\n'.join(lines)
            if return_str:
                return msg
            logger.info(msg)

        
class validate_yaml(SubCommand):
    r"""Validate a set of or or more YAMLs defining an integration."""

    name = "validate"
    help = 'Validate a set of YAML specification files for an integration.'
    arguments = [
        (('yamlfile', ),
         {'nargs': '+',
          'help': 'One or more YAML specification files.'})]

    @classmethod
    def func(cls, args):
        from yggdrasil import yamlfile
        yamlfile.parse_yaml(args.yamlfile)
        logger.info("Validation succesful.")


class yggcc(SubCommand):
    r"""Compile a program."""

    name = "compile"
    help = ("Compile a program from source files for use in an "
            "yggdrasil integration.")
    arguments = [
        (('source', ),
         {'nargs': '+',
          'help': ("One or more source files or the directory containing an "
                   "R package.")}),
        (('--language', ),
         {'default': None,
          'choices': [None] + LANGUAGES_WITH_ALIASES['compiled'] + ['R', 'r'],
          'help': ("Language of the source code. If not provided, "
                   "the language will be determined from the "
                   "source extension.")}),
        (('--toolname', ),
         {'help': "Name of compilation tool that should be used"}),
        (('--flags', ),
         {'nargs': '*',
          'help': ("Additional flags that should be added to the compilation "
                   "command")}),
        (('--use-ccache', ),
         {'action': 'store_true', 'help': "Run compilation with ccache."}),
        (('--Rpkg-language', ),
         {'help': ("Language that R package is written in "
                   "(only used if the specified language is R).")})]

    @classmethod
    def func(cls, args):
        from yggdrasil.components import import_component
        from yggdrasil.constants import EXT2LANG
        if args.language is None:
            if (((len(args.source) == 1) and os.path.isdir(args.source[0])
                 and os.path.isdir(os.path.join(args.source[0], 'R')))):
                args.language = 'R'
            else:
                args.language = EXT2LANG[os.path.splitext(args.source[0])[-1]]
        drv = import_component('model', args.language)
        kws = {'toolname': args.toolname, 'flags': args.flags,
               'use_ccache': args.use_ccache}
        if (args.language in ['r', 'R']) and args.Rpkg_language:
            kws['language'] = args.Rpkg_language
        print("executable: %s" % drv.call_compiler(args.source, **kws))


class yggcompile(SubCommand):
    r"""Compile interface library/libraries."""

    name = "compile-deps"
    help = ("Compile yggdrasil dependency libraries. Existing "
            "libraries are first deleted.")
    arguments = [
        (('language', ),
         {'nargs': '*', 'default': ['all'],
          # 'choices': (['all'] + LANGUAGES_WITH_ALIASES['compiled']
          #             + LANGUAGES_WITH_ALIASES['compiled_dsl']),
          'help': ("One or more languages to compile dependencies "
                   "for.")}),
        (('--toolname', ),
         {'help': "Name of compilation tool that should be used"})]

    @classmethod
    def func(cls, args):
        from yggdrasil.components import import_component
        yggclean.func(args, verbose=False)
        error_on_missing = (not getattr(args, 'all_languages', False))
        missing = []
        for lang in args.language:
            import_component('model', lang).cleanup_dependencies()
            drv = import_component('model', lang)
            if ((hasattr(drv, 'compile_dependencies')
                 and (not getattr(drv, 'is_build_tool', False)))):
                if drv.is_installed():
                    drv.compile_dependencies(toolname=args.toolname)
                else:
                    missing.append(lang)
        if error_on_missing and missing:  # pragma: debug
            raise Exception(("One or more of the requested languages "
                             "are not fully installed for use with "
                             "yggdrasil: %s") % missing)


class yggclean(SubCommand):
    r"""Cleanup dependency files."""

    name = "clean"
    help = "Remove dependency libraries compiled by yggdrasil."
    arguments = [
        (('language', ),
         {'nargs': '*', 'default': ['all'],
          # 'choices': ['all'] + LANGUAGES_WITH_ALIASES['all'],
          'help': ("One or more languages to clean up dependencies "
                   "for.")})]

    @classmethod
    def func(cls, args, verbose=True):
        from yggdrasil.components import import_component
        for lang in args.language:
            if lang in ['ipc', 'ipcs']:
                from yggdrasil.communication.IPCComm import ipcrm_queues
                ipcrm_queues()
            else:
                import_component('model', lang).cleanup_dependencies(
                    verbose=verbose)


class cc_toolname(SubCommand):
    r"""Output the name of the compiler used to compile C or C++ programs."""

    name = "compiler-tool"
    help = 'Get the compiler used for C/C++ programs.'
    arguments = [
        (('--cpp', ),
         {'action': 'store_true',
          'help': 'Get the compiler used for C++ programs.'}),
        (('--fullpath', ),
         {'action': 'store_true',
          'help': 'Get the full path to the tool exectuable.'})]

    @classmethod
    def parse_args(cls, *args, **kwargs):
        args = super(cc_toolname, cls).parse_args(*args, **kwargs)
        args.tool = 'compiler'
        args.toolname = None
        args.flags = False
        if args.cpp:
            args.language = 'cpp'
        else:
            args.language = 'c'
        return args

    @classmethod
    def func(cls, args, **kwargs):
        ygginfo.func(args, **kwargs)


class ld_toolname(cc_toolname):
    r"""Output the name of the linker used to compile C or C++ programs."""

    name = "linker-tool"
    help = 'Get the linker used for C/C++ programs.'
    arguments = [
        (('--cpp', ),
         {'action': 'store_true',
          'help': 'Get the linker used for C++ programs.'}),
        (('--fullpath', ),
         {'action': 'store_true',
          'help': 'Get the full path to the tool exectuable.'})]

    @classmethod
    def parse_args(cls, *args, **kwargs):
        args = super(ld_toolname, cls).parse_args(*args, **kwargs)
        args.tool = 'linker'
        return args


class cc_flags(cc_toolname):
    r"""Get the compiler flags necessary for including the interface
    library in a C or C++ program."""

    name = "compiler-flags"
    help = 'Get the compilation flags necessary for a C/C++ program.'
    arguments = [
        (('--cpp', ),
         {'action': 'store_true',
          'help': 'Get the compilation flags used for C++ programs.'}),
        (('--toolname', ),
         {'default': None,
          'help': 'Name of the tool that associated flags be returned for.'})]

    @classmethod
    def parse_args(cls, *args, **kwargs):
        args = super(cc_flags, cls).parse_args(*args, **kwargs)
        args.flags = True
        return args


class ld_flags(cc_toolname):
    r"""Get the linker flags necessary for including the interface
    library in a C or C++ program."""

    name = "linker-flags"
    help = 'Get the compilation flags necessary for a C/C++ program.'
    arguments = [
        (('--cpp', ),
         {'action': 'store_true',
          'help': 'Get the compilation flags used for C++ programs.'}),
        (('--toolname', ),
         {'default': None,
          'help': 'Name of the tool that associated flags be returned for.'})]

    @classmethod
    def parse_args(cls, *args, **kwargs):
        args = super(ld_flags, cls).parse_args(*args, **kwargs)
        args.tool = 'linker'
        args.flags = True
        return args


class ygginstall(SubCommand):
    r"""Call installation script."""

    name = "install"

    @classmethod
    def get_parser(cls, **kwargs):
        from yggdrasil.languages import install_languages
        # TODO: Migrate
        return install_languages.update_argparser()

    @classmethod
    def func(cls, args):
        from yggdrasil.languages import install_languages
        for x in args.language:
            install_languages.install_language(x, args=args)


class update_config(SubCommand):
    r"""Update the user config file for yggdrasil."""

    name = "config"
    help = 'Update the user config file.'
    arguments = (
        [(('languages', ),
          {'nargs': '*',
           # 'choices': ['all'] + LANGUAGES_WITH_ALIASES['all'],
           'default': [],
           'help': 'One or more languages that should be configured.'}),
         (('--languages', ),
          {'nargs': '+', 'dest': 'languages_flag',
           # 'choices': ['all'] + LANGUAGES_WITH_ALIASES['all'],
           'default': [],
           'help': 'One or more languages that should be configured.'}),
         (('--show-file', ),
          {'action': 'store_true',
           'help': 'Print the path to the config file without updating it.'}),
         (('--remove-file', ),
          {'action': 'store_true',
           'help': 'Remove the existing config file and return.'}),
         (('--overwrite', ),
          {'action': 'store_true',
           'help': 'Overwrite the existing file.'}),
         (('--disable-languages', ),
          {'nargs': '+', 'default': [],
           'choices': LANGUAGES_WITH_ALIASES['all'],
           'help': 'One or more languages that should be disabled.'}),
         (('--enable-languages', ),
          {'nargs': '+', 'default': [],
           'choices': LANGUAGES_WITH_ALIASES['all'],
           'help': 'One or more languages that should be enabled.'}),
         (('--quiet', '-q'),
          {'action': 'store_true',
           'help': 'Suppress output.'}),
         (('--allow-multiple-omp', ),
          {'action': 'store_true', 'default': None,
           'help': ('Have yggdrasil set the environment variable '
                    'KMP_DUPLICATE_LIB_OK to \'True\' during model runs '
                    'to disable errors resembling '
                    '"OMP: Error #15: Initializing libomp.dylib..." '
                    'that result from having multiple versions of OpenMP '
                    'loaded during runtime.')}),
         (('--dont-allow-multiple-omp', ),
          {'action': 'store_true', 'default': None,
           'help': ('Don\'t set the KMP_DUPLICATE_LIB_OK environment variable '
                    'when running models (see help for \'--allow-multiple-omp\' '
                    'for more information).')})]
        + [(('--%s-compiler' % k, ),
            {'help': ('Name or path to compiler that should be used to compile '
                      'models written in %s.' % k)})
           for k in LANGUAGES['compiled']]
        + [(('--%s-linker' % k, ),
            {'help': ('Name or path to linker that should be used to link '
                      'models written in %s.' % k)})
           for k in LANGUAGES['compiled']]
        + [(('--%s-archiver' % k, ),
            {'help': ('Name or path to archiver that should be used to create '
                      'static libraries for models written in %s.' % k)})
           for k in LANGUAGES['compiled']]
    )
    # TODO: Move these into the language directories?
    language_arguments = {
        'c': [
            ConditionalArgumentTuple(
                ('--vcpkg-dir', ),
                {'help': 'Directory containing the vcpkg installation.'},
                conditions={'os': ['Windows']}),
            ConditionalArgumentTuple(
                ('--macos-sdkroot', '--sdkroot'),
                {'help': ('The full path to the MacOS SDK that '
                          'should be used.')},
                conditions={'os': ['MacOS']})],
        'matlab': [
            (('--disable-matlab-engine-for-python', ),
             {'action': 'store_true', 'default': None,
              'dest': 'disable_engine',
              'help': 'Disable use of the Matlab engine for Python.'}),
            (('--enable-matlab-engine-for-python', ),
             {'action': 'store_true', 'default': None,
              'dest': 'enable_engine',
              'help': 'Enable use of the Matlab engine for Python.'}),
            (('--hide-matlab-libiomp', ),
             {'action': 'store_true', 'default': None,
              'help': ('Hide the version of libiomp installed by Matlab '
                       'by slightly changing the filename so that the '
                       'conda version of libomp is used instead. This '
                       'helps to solve the error "'
                       'OMP: Error #15: Initializing libomp.dylib..." '
                       'that can occur when using a conda environment. '
                       'The hidden file location will be set in the '
                       'configuration file and can be restored via the '
                       '\'--restore-matlab-libiomp\' option.')}),
            (('--restore-matlab-libiomp', ),
             {'action': 'store_true', 'default': None,
              'help': ('Restore the version of libiomp installed by Matlab. '
                       '(See help for \'--hide-matlab-libiomp\')')}),
        ],
        'osr': [
            (('--osr-repository-path', ),
             {'dest': 'repository',
              'help': 'The full path to the OpenSimRoot repository.'})]}
    opposite_arguments = [
        ('allow_multiple_omp', 'dont_allow_multiple_omp'),
        ('disable_engine', 'enable_engine'),
        ('hide_matlab_libiomp', 'restore_matlab_libiomp')]
        
    @classmethod
    def add_arguments(cls, parser, **kwargs):
        super(update_config, cls).add_arguments(parser, **kwargs)
        args = kwargs.get('args', None)
        if args is None:
            args = sys.argv[1:]
        if ('-h' in args) or ('--help' in args):
            args = [x for x in args if x not in ['-h', '--help']]
        preargs = parser.parse_known_args(args=args)[0]
        prelang = preargs.languages
        if preargs.languages_flag:
            prelang += preargs.languages_flag
        if (len(prelang) == 0) or ('all' in prelang):
            prelang = LANGUAGES['all']
        # TODO: The languages could be subparsers
        for k, v in cls.language_arguments.items():
            if k in prelang:
                cls.add_argument_to_parser(parser, v)

    @classmethod
    def func(cls, args):
        from yggdrasil import config
        if args.show_file:
            print('Config file located here: %s'
                  % config.usr_config_file)
        if args.remove_file and os.path.isfile(config.usr_config_file):
            os.remove(config.usr_config_file)
        if args.show_file or args.remove_file:
            return
        for x_true, x_false in cls.opposite_arguments:
            if getattr(args, x_false, None) is not None:
                assert(getattr(args, x_true, None) is None)
            setattr(args, x_true, not getattr(args, x_false))
            delattr(args, x_false)
        lang_kwargs = {}
        for k, v in cls.language_arguments.items():
            for v_args in v:
                name = v_args[1].get(
                    'dest',
                    v_args[0][0].lstrip('-').replace('-', '_'))
                if getattr(args, name, None) is not None:
                    lang_kwargs.setdefault(k, {})
                    lang_kwargs[k][name] = getattr(args, name)
        for x in ['compiler', 'linker', 'archiver']:
            for k in LANGUAGES['compiled']:
                if getattr(args, '%s_%s' % (k, x), None):
                    lang_kwargs.setdefault(k, {})
                    lang_kwargs[k][x] = getattr(args, '%s_%s' % (k, x))
        config.update_language_config(
            args.languages, overwrite=args.overwrite,
            verbose=(not args.quiet),
            disable_languages=args.disable_languages,
            enable_languages=args.enable_languages,
            allow_multiple_omp=args.allow_multiple_omp,
            lang_kwargs=lang_kwargs)


class regen_metaschema(SubCommand):
    r"""Regenerate the yggdrasil metaschema."""

    name = "metaschema"
    help = "Regenerate the yggdrasil metaschema."

    @classmethod
    def func(cls, args):
        from yggdrasil import metaschema
        if os.path.isfile(metaschema._metaschema_fname):
            os.remove(metaschema._metaschema_fname)
        metaschema._metaschema = None
        metaschema._validator = None
        metaschema.get_metaschema()


class regen_schema(SubCommand):
    r"""Regenerate the yggdrasil schema."""

    name = "schema"
    help = "Regenerate the yggdrasil schema."
    arguments = [
        (('--only-constants', ),
         {'action': 'store_true',
          'help': ('Only update the constants.py file without updating '
                   'the schema.')})]

    @classmethod
    def func(cls, args):
        from yggdrasil import schema
        if not args.only_constants:
            if os.path.isfile(schema._schema_fname):
                os.remove(schema._schema_fname)
            schema.clear_schema()
            schema.init_schema()
        schema.update_constants()


class yggmodelform(SubCommand):
    r"""Save/print a JSON schema that can be used for generating a
    form for composing a model specification files."""

    name = "model-form-schema"
    help = ('Save/print the JSON schema for generating the model '
            'specification form.')
    arguments = [
        (('--file', ),
         {'help': 'Path to file where the schema should be saved.'})]

    @classmethod
    def func(cls, args):
        from yggdrasil.schema import get_model_form_schema
        out = get_model_form_schema(fname_dst=args.file)
        if not args.file:
            pprint.pprint(out)


class yggdevup(SubCommand):
    r"""Cleanup old libraries, re-install languages, and re-compile
    interface libraries following an update to the code (when doing
    development)."""

    name = "dev-update"
    help = ('Perform cleanup and reinitialization following an '
            'update to the code during development.')

    @classmethod
    def func(cls, args):
        yggclean(args=['all'])
        ygginstall(args=['all'])
        yggcompile(args=['all'])
        ygginfo(args=[])


class run_tsts(SubCommand):
    r"""Run yggdrasil test suite."""
    # TODO: Move this into a pytest extension?

    name = "test"
    help = 'Run yggdrasil tests.'
    arguments = [
        (('--withcoverage', '--with-coverage'),
         {'action': 'store_true',
          'help': 'Record coverage during tests.'}),
        (('--verbose', '-v'),
         {'action': 'store_true',
          'help': 'Increase verbosity of output from the test runner'}),
        (('--nocapture', '-s'),
         {'action': 'store_true',
          'help': 'Don\'t capture output from tests.'}),
        (('--stop', '-x'),
         {'action': 'store_true',
          'help': 'Stop after first test failure.'}),
        (('--nologcapture', ),
         {'action': 'store_true',
          'help': ('Don\'t capture output from log messages '
                   'generated during tests.')}),
        (('--noflaky', '--no-flaky'),
         {'action': 'store_true',
          'help': 'Don\'t re-run flaky tests.'}),
        (('--ci', ),
         {'action': 'store_true',
          'help': ('Perform addition operations required for '
                   'testing on continuous integration services.')}),
        (('--test-suite', '--test-suites'),
         {'nargs': '+', 'action': 'extend', 'type': str,
          'choices': ['all', 'top', 'examples', 'examples_part1',
                      'examples_part2', 'demos', 'types', 'timing',
                      'connections', 'models'],
          'help': 'Test suite(s) that should be run.',
          'dest': 'test_suites'}),
        (('--pytest-config', '-c'),
         {'help': ('Pytest configuration file that should be used '
                   'instead of the defaults.')}),
        (('--cov-config', ),
         {'help': 'Config file for coverage. Default: .coveragerc'}),
        (('--ignore', ),
         {'action': 'append',
          'help': 'Ignore path during collection.'}),
        (('--rootdir', ),
         {'help': 'Define root directory for tests.'}),
        (('--count', ),
         {'type': int, 'default': 1,
          'help': 'Number of times to repeat a test.'}),
        (('--additional-info', '-r'),
         {'type': str, 'default': '',
          'help': 'Display additional info for test results.'})]
    allow_unknown = True

    @classmethod
    def add_arguments(cls, parser, **kwargs):
        from yggdrasil import config
        super(run_tsts, cls).add_arguments(parser, **kwargs)
        parser = config.get_config_parser(parser)

    @classmethod
    def parse_args(cls, *args, **kwargs):
        from yggdrasil import config, platform
        args = super(run_tsts, cls).parse_args(*args, **kwargs)
        extra = args._extra_commands
        if args.ci:
            args.verbose = True
            args.withcoverage = True
            args.setup_cfg = 'setup.cfg'
            args.pytest_config = args.setup_cfg
            args.cov_config = '.coveragerc'
            if not args.ignore:
                args.ignore = []
            args.ignore.append('yggdrasil/rapidjson/')
            args.rootdir = package_dir
            extra += ['--reruns=2', '--reruns-delay=1', '--timeout=900']
        # Separate out paths from options
        test_paths = []
        if args.test_suites:
            if 'all' in args.test_suites:
                args.test_suites.remove('all')
                for x in ['top', 'examples', 'demos', 'types', 'timing',
                          'connections', 'models']:
                    if x not in args.test_suites:
                        args.test_suites.append(x)
            for x in args.test_suites:
                if x == 'top':
                    test_paths.append(package_dir)
                elif x == 'examples':
                    args.enable_examples = True
                    test_paths.append('examples')
                elif x == 'examples_part1':
                    args.enable_examples = True
                    if platform._is_win:  # pragma: windows
                        pattern = 'test_[a-g]*.py'
                    else:
                        pattern = 'test_[A-Za-g]*.py'
                    test_paths.append(os.path.join(
                        'examples', 'tests', pattern))
                elif x == 'examples_part2':
                    args.enable_examples = True
                    pattern = 'test_[h-z]*.py'
                    test_paths.append(os.path.join(
                        'examples', 'tests', pattern))
                elif x == 'demos':
                    args.enable_demos = True
                    test_paths.append('demos')
                elif x == 'types':
                    args.enable_examples = True
                    args.long_running = True
                    test_paths.append(os.path.join('examples', 'tests',
                                                   'test_types.py'))
                elif x == 'timing':
                    args.long_running = True
                    args.enable_production_runs = True
                    test_paths.append(os.path.join('tests', 'test_timing.py'))
                elif x == 'connections':
                    for f in ['test_ConnectionDriver.py',
                              'test_FileInputDriver.py',
                              'test_FileOutputDriver.py',
                              'test_RPCRequestDriver.py']:
                        test_paths.append(
                            os.path.join('drivers', 'tests', f))
                elif x == 'models':
                    test_paths.append(
                        os.path.join('drivers', 'tests',
                                     'test_*ModelDriver.py'))
        if (not test_paths) and all(x.startswith('-') for x in extra):
            test_paths.append(package_dir)
        # Get expanded tests to allow for paths that are relative to
        # either the yggdrasil root directory or the current working
        # directory
        args = config.resolve_config_parser(args)
        args.extra = []
        cls.expand_and_add(extra + test_paths, args.extra,
                           [package_dir, os.getcwd()],
                           add_on_nomatch=True)
        return args

    @classmethod
    def expand_and_add(cls, path, path_list, dir_list,
                       add_on_nomatch=False):
        r"""Expand the specified path and add it's expanded forms to
        the provided list.

        Args:
            path (str): Absolute/relative path with or without regex
                wildcard expressions to expand.
            path_list (list): Existing list that expansions should be
                added to.
            dir_list (list): Directories that should be tried for
                relative paths.
            add_on_nomatch (bool, optional): If True, path will be added
                to path_list even if there are not matches. Defaults to
                False.

        Returns:
            int: The number of expansions added to path_list.

        """
        import glob
        if isinstance(path, list):
            nadded = 0
            for x in path:
                nadded += cls.expand_and_add(x, path_list, dir_list,
                                             add_on_nomatch=add_on_nomatch)
            return nadded
        if path.startswith('-'):
            if add_on_nomatch:
                path_list.append(path)
                return 1
            return 0
        elif path.endswith('yggtest'):
            return 0
        elif os.path.isabs(path):
            matches = sorted(glob.glob(path))
            path_list += matches
            if (not matches) and add_on_nomatch:
                path_list.append(path)
                return 1
            return len(matches)
        # Try checking for function
        for func_sep in ['::', ':']:
            if (func_sep in path):
                prepath, mod = path.rsplit(func_sep, 1)
                nadded = cls.expand_and_add(prepath, path_list, dir_list)
                if nadded:
                    for i in range(-nadded, 0):
                        path_list[i] += '%s%s' % (func_sep, mod)
                    return nadded
        # Try directory prefixes
        for path_prefix in dir_list:
            nadded = cls.expand_and_add(os.path.join(path_prefix, path),
                                        path_list, dir_list)
            if nadded:
                return nadded
        if add_on_nomatch:
            path_list.append(path)
            return 1
        return 0

    @classmethod
    def func(cls, args):
        from yggdrasil import config
        argv = [sys.executable, '-m', 'pytest']
        # test_paths = args.test_paths
        if args.verbose:
            argv.append('-v')
        if args.nocapture:
            argv.append('-s')
            argv += ['-o', 'log_cli=true']
        if args.nologcapture:
            argv += ['-o', 'log_cli=true']
        if args.stop:
            argv.append('-x')
        if args.withcoverage:
            # See information about getting coverage of test fixtures
            # https://pytest-cov.readthedocs.io/en/stable/plugins.html
            argv.append('--cov=%s' % package_dir)
            # argv.append('--cov-append')
        if args.noflaky:
            argv += ['-p', 'no:flaky']
        if args.pytest_config:
            argv += ['-c', args.pytest_config]
        if args.cov_config:
            argv += ['--cov-config=%s' % args.cov_config]
        if args.ignore:
            argv += ['--ignore=%s' % x for x in args.ignore]
        if args.rootdir:
            argv += ['--rootdir=%s' % args.rootdir]
        if args.additional_info:
            argv += ['-r', args.additional_info]
        argv += args.extra
        # Run test command and perform cleanup before logging any errors
        logger.info("Running %s from %s", argv, os.getcwd())
        new_config = {}
        # pth_file = 'ygg_coverage.pth'
        # assert(not os.path.isfile(pth_file))
        # # Get arguments for temporary test environment
        # if args.withcoverage:
        #     new_config['COVERAGE_PROCESS_START'] = 'True'
        #     with open(pth_file, 'w') as fd:
        #         fd.write("import coverage; coverage.process_startup()")
        initial_dir = os.getcwd()
        error_code = 0
        with config.parser_config(args, **new_config):
            try:
                # Perform CI specific pretest operations
                if args.ci:
                    if not os.path.isfile(args.setup_cfg):
                        raise RuntimeError("The CI tests must be run "
                                           "from the root directory "
                                           "of the yggdrasil git "
                                           "repository.")
                    top_dir = os.path.dirname(os.getcwd())
                    src_cmd = ('python -c \"import versioneer; '
                               'print(versioneer.get_version())\"')
                    dst_cmd = ('python -c \"import yggdrasil; '
                               'print(yggdrasil.__version__)\"')
                    src_ver = subprocess.check_output(src_cmd,
                                                      shell=True)
                    dst_ver = subprocess.check_output(dst_cmd,
                                                      shell=True,
                                                      cwd=top_dir)
                    if src_ver != dst_ver:  # pragma: debug
                        raise RuntimeError(("Versions do not match:\n"
                                            "\tSource version: %s\n"
                                            "\tBuild  version: %s\n")
                                           % (src_ver, dst_ver))
                    subprocess.check_call(
                        ["flake8", "yggdrasil", "--append-config",
                         args.setup_cfg])
                    if os.environ.get("YGG_CONDA", None):
                        subprocess.check_call(
                            ["python", "create_coveragerc.py"])
                    if not os.path.isfile(".coveragerc"):
                        raise RuntimeError(".coveragerc file dosn't "
                                           "exist.")
                    with open(".coveragerc", "r") as fd:
                        print(fd.read())
                    subprocess.check_call(["yggdrasil", "info",
                                           "--verbose"])
                error_code = 0
                for x in range(args.count):
                    x_error_code = subprocess.call(argv)
                    if x_error_code != 0:
                        error_code = x_error_code
                        if args.stop:
                            break
            # except BaseException:
            #     logger.exception('Error in running test.')
            #     error_code = -1
            finally:
                os.chdir(initial_dir)
                # if os.path.isfile(pth_file):
                #     os.remove(pth_file)
        return error_code


class timing_plots(SubCommand):
    r"""Create performance plots using timing results."""

    name = "timing"
    help = "Create performance plots using timing results."
    arguments = [
        ArgumentSubparser(
            title='comparison', dest='comparison',
            description='Comparison plot that should be created.',
            parsers=[
                ArgumentParser(
                    name='commtype',
                    help='Compare different communication avenues.'),
                ArgumentParser(
                    name='language',
                    help='Compare different programming languages.'),
                ArgumentParser(
                    name='os',
                    help='Compare different operating systems.'),
                ArgumentParser(
                    name='python',
                    help='Compare different versions of Python.'),
                ArgumentParser(
                    name='lang2019',
                    help='Create plots from Lang (2019) paper.')])]

    @classmethod
    def func(cls, args):
        from yggdrasil import timing
        if args.comparison == 'lang2019':
            _lang_list = timing._lang_list
            _lang_list_nomatlab = copy.deepcopy(_lang_list)
            _lang_list_nomatlab.remove('matlab')
            timing.plot_scalings(compare='platform', python_ver='2.7')
            # All plots on Linux, no matlab
            timing.plot_scalings(compare='comm_type',
                                 platform='Linux',
                                 python_ver='2.7')
            timing.plot_scalings(compare='python_ver',
                                 platform='Linux')
            timing.plot_scalings(compare='language',
                                 platform='Linux',
                                 python_ver='2.7',
                                 compare_values=_lang_list_nomatlab)
            # Language comparision on MacOS, with matlab
            timing.plot_scalings(compare='language',
                                 platform='MacOS',
                                 python_ver='2.7',
                                 compare_values=_lang_list)
        else:
            if args.comparison == 'commtype':
                args.comparison = 'comm_type'
            elif args.comparison == 'os':
                args.comparison = 'platform'
            timing.plot_scalings(compare=args.comparison)


class generate_gha_workflow(SubCommand):
    r"""Re-generate the Github actions workflow yaml."""

    name = "gha"
    help = (
        "Generate a Github Actions (GHA) workflow yaml file from "
        "a version of the file that uses anchors (not supported by "
        "GHA as of 2021-01-14).")
    arguments = [
        (('--base', '--base-file'),
         {'help': (
             "Version of GHA workflow yaml that contains anchors.")}),
        (('--dest', ),
         {'help': (
             "Name of target GHA workflow yaml file.")}),
        (('--verbose', ),
         {'action': 'store_true', 'help': "Print yaml contents."})]

    @classmethod
    def func(cls, args, gitdir=None):
        import yaml
        from yggdrasil.metaschema.encoder import decode_yaml, encode_yaml
        from collections import OrderedDict

        class NoAliasDumper(yaml.SafeDumper):
            def ignore_aliases(self, data):
                return True
        if gitdir is None:
            try:
                gitdir = subprocess.check_output(
                    ["git", "rev-parse", "--git-dir"],
                    stderr=subprocess.PIPE).decode('utf-8').strip()
            except subprocess.CalledProcessError:
                return 1
        if args.base is None:
            args.base = os.path.join(gitdir, '..', 'utils',
                                     'test-install-base.yml')
        if args.dest is None:
            args.dest = os.path.join(gitdir, '..', '.github', 'workflows',
                                     'test-install.yml')
        with open(args.base, 'r') as fd:
            contents = decode_yaml(fd.read(), sorted_dict_type=OrderedDict)
            # contents = yaml.load(fd, Loader=yaml.SafeLoader)
        if args.verbose:
            pprint.pprint(contents)
        with open(args.dest, 'w') as fd:
            fd.write(('# DO NOT MODIFY THIS FILE, IT IS GENERATED.\n'
                      '# To make changes, modify \'%s\'\n'
                      '# and run \'ygggha\'\n')
                     % args.base)
            fd.write(encode_yaml(contents, sorted_dict_type=OrderedDict,
                                 Dumper=NoAliasDumper))
            # yaml.dump(contents, fd, Dumper=NoAliasDumper)


def rebuild_c_api():
    r"""Rebuild the C/C++ API."""
    ReplacementWarning('yggbuildapi_c', 'yggdrasil compile-deps c cpp')
    yggcompile(args=['c', 'cpp'] + sys.argv[1:])


def yggtime_comm():
    r"""Plot timing statistics comparing the different communication mechanisms."""
    ReplacementWarning('yggtime_comm', 'yggdrasil timing commtype')
    timing_plots(args=['commtype'] + sys.argv[1:])


def yggtime_lang():
    r"""Plot timing statistics comparing the different languages."""
    ReplacementWarning('yggtime_lang', 'yggdrasil timing language')
    timing_plots(args=['language'] + sys.argv[1:])


def yggtime_os():
    r"""Plot timing statistics comparing the different operating systems."""
    ReplacementWarning('yggtime_os', 'yggdrasil timing os')
    timing_plots(args=['os'] + sys.argv[1:])


def yggtime_py():
    r"""Plot timing statistics comparing the different versions of Python."""
    ReplacementWarning('yggtime_py', 'yggdrasil timing python')
    timing_plots(args=['python'] + sys.argv[1:])


def yggtime_paper():
    r"""Create plots for timing."""
    ReplacementWarning('yggtime_paper', 'yggdrasil timing lang2019')
    timing_plots(args=['lang2019'] + sys.argv[1:])


if __name__ == '__main__':
    yggrun()
    sys.exit(0)



import numpy as np
from yggdrasil import units


def light(doy, height):
    r"""Compute the intensity of light.

    Args:
       doy (float): Day of year.
       height (float): Distance from ground in cm.

    Returns:
       float: Intensity of light in ergs cm^-2 s^-1.

    """
    # Define parameters that are static across a run
    amplitude = units.add_units(80.0, 'ergs cm^-3 s^-1')
    doy_offset = units.add_units(0.0, 'days')

    # Calculate intensity
    intensity = (
        amplitude * height *
        (1.0 + np.sin(2.0 * np.pi * (doy - doy_offset) /
                      units.add_units(365.0, 'days'))))
    
    return intensity



from yggdrasil.tests import generate_component_tests
from yggdrasil.communication.transforms.tests.test_TransformBase import TestTransformBase


generate_component_tests('transform', TestTransformBase,
                         globals(), __file__, class_attr='transform')


__all__ = []



import sys
from yggdrasil import units
from yggdrasil.interface.YggInterface import (
    YggTimesync, YggOutput)


def timestep_calc(t):
    r"""Updates the state based on the time where x is a sine wave
    with period of 10 days and y is a cosine wave with a period of 5 days.

    Args:
        t (float): Current time.

    Returns:
        dict: Map of state parameters.

    """
    state = {"carbonAllocation2Roots": units.add_units(10.0, 'g'),
             "saturatedConductivity": units.add_units(10.0, 'cm/day')}
    return state


def main(t_step, t_units):
    r"""Function to execute integration.

    Args:
        t_step (float): The time step that should be used.
        t_units (str): Units of the time step.

    """
    print('Hello from Python other_model: timestep = %s %s' % (t_step, t_units))
    t_step = units.add_units(t_step, t_units)
    t_start = units.add_units(0.0, t_units)
    t_end = units.add_units(1.0, 'day')
    state = timestep_calc(t_start)

    # Set up connections matching yaml
    # Timestep synchonization connection will default to 'timesync'
    timesync = YggTimesync('timesync')
    out = YggOutput('output')

    # Initialize state and synchronize with other models
    t = t_start
    ret, state = timesync.call(t, state)
    if not ret:
        raise RuntimeError("other_model(Python): Initial sync failed.")
    print('other_model(Python): t = % 8s' % t, end='')
    for k, v in state.items():
        print(', %s = %+ 5.2f' % (k, v), end='')
    print('')

    # Send initial state to output
    flag = out.send(dict(state, time=t))
    if not flag:
        raise RuntimeError("other_model(Python): Failed to send "
                           "initial output for t=%s." % t)
    
    # Iterate until end
    while t < t_end:

        # Perform calculations to update the state
        t = t + t_step
        state = timestep_calc(t)

        # Synchronize the state
        ret, state = timesync.call(t, state)
        if not ret:
            raise RuntimeError("other_model(Python): sync for t=%f failed." % t)
        print('other_model(Python): t = % 8s' % t, end='')
        for k, v in state.items():
            print(', %s = %+ 5.2f' % (k, v), end='')
        print('')

        # Send output
        flag = out.send(dict(state, time=t))
        if not flag:
            raise RuntimeError("other_model(Python): Failed to send output for t=%s." % t)

    print('Goodbye from Python other_model')


if __name__ == '__main__':
    # Take time step from the first argument
    main(float(sys.argv[1]), sys.argv[2])



from yggdrasil.communication.transforms.TransformBase import TransformBase


class FilterTransform(TransformBase):
    r"""Class for applying a filter."""
    _transformtype = 'filter'
    _schema_required = ['filter']
    _schema_properties = {'filter': {'$ref': '#/definitions/filter'}}

    def __init__(self, *args, **kwargs):
        super(FilterTransform, self).__init__(*args, **kwargs)
        if isinstance(self.filter, dict):
            from yggdrasil.schema import get_schema
            from yggdrasil.components import create_component
            filter_schema = get_schema().get('filter')
            filter_kws = dict(self.filter,
                              subtype=filter_schema.identify_subtype(self.filter))
            self.filter = create_component('filter', **filter_kws)
    
    def transform_datatype(self, datatype):
        r"""Determine the datatype that will result from applying the transform
        to the supplied datatype.

        Args:
            datatype (dict): Datatype to transform.

        Returns:
            dict: Transformed datatype.

        """
        # Filter should not actually modify the message
        return datatype

    def evaluate_transform(self, x, no_copy=False):
        r"""Call transform on the provided message.

        Args:
            x (object): Message object to transform.
            no_copy (bool, optional): If True, the transformation occurs in
                place. Otherwise a copy is created and transformed. Defaults
                to False.

        Returns:
            object: The transformed message.

        """
        if self.filter(x):
            return x
        return iter([])

    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the transform class.

        Returns:
            list: Multiple dictionaries of keywords and messages before/after
                pairs that will result from the transform created by the provided
                keywords.
        
        """
        return [
            {'kwargs': {'filter': {'statement': '%x% > 1'}},
             'in/out': [(0, iter([])),
                        (2, 2)],
             'in/out_t': [({'type': 'int'}, {'type': 'int'})]}
        ]



import os
import json
import shutil
from yggdrasil.tests import assert_equal, assert_raises
from yggdrasil import config, tools, platform


def test_YggConfigParser():
    r"""Ensure that get returns proper defaults etc."""
    x = config.YggConfigParser()
    x.add_section('test_section')
    x.set('test_section', 'test_option', 'test_value')
    assert_raises(RuntimeError, x.update_file)
    assert_equal(x.file_to_update, None)
    assert_equal(x.get('test_section', 'test_option'), 'test_value')
    assert_equal(x.get('test_section', 'fake_option'), None)
    assert_equal(x.get('test_section', 'fake_option', 5), 5)
    assert_equal(x.get('fake_section', 'fake_option'), None)
    assert_equal(x.get('fake_section', 'fake_option', 5), 5)


def test_update_language_config():
    r"""Test updating configuration for installed languages."""
    languages = tools.get_supported_lang()
    cfg_orig = config.ygg_cfg_usr.file_to_update
    cfg_copy = '_copy'.join(os.path.splitext(cfg_orig))
    shutil.copy2(cfg_orig, cfg_copy)
    try:
        if config.ygg_cfg_usr.has_option('c', 'vcpkg_dir'):
            config.update_language_config(
                'c', lang_kwargs={'c': {
                    'vcpkg_dir': config.ygg_cfg_usr.get(
                        'c', 'vcpkg_dir')}})
        if platform._is_mac:
            config.update_language_config(
                'c', lang_kwargs={'c': {'compiler': shutil.which('clang')}})
        config.update_language_config(overwrite=True, verbose=True,
                                      allow_multiple_omp=True)
        if len(languages) > 0:
            config.ygg_cfg_usr.remove_section(languages[0])
            config.update_language_config(
                languages[0], disable_languages=[languages[0]])
            config.update_language_config(
                languages[0], disable_languages=[languages[0]],
                enable_languages=[languages[0]])
            config.update_language_config(
                languages[0], enable_languages=[languages[0]])
        if len(languages) > 1:
            config.ygg_cfg_usr.remove_section(languages[1])
            config.update_language_config(
                languages[1], enable_languages=[languages[1]])
    finally:
        shutil.move(cfg_copy, cfg_orig)
        config.ygg_cfg_usr.reload()
        config.ygg_cfg.reload()


def test_cfg_logging():
    r"""Test cfg_logging."""
    lvl = config.get_ygg_loglevel()
    config.set_ygg_loglevel(lvl)
    os.environ['YGG_SUBPROCESS'] = 'True'
    lvl = config.get_ygg_loglevel()
    config.set_ygg_loglevel(lvl)
    config.cfg_logging()
    del os.environ['YGG_SUBPROCESS']


def test_parser_config():
    r"""Test parser_config."""
    config.get_config_parser(description="test", skip_sections=['testing'])
    parser = config.get_config_parser(description="test")
    assert_raises(ValueError, config.resolve_config_parser,
                  parser.parse_args(['--production-run', '--debug']))
    args1 = config.resolve_config_parser(
        parser.parse_args(['--languages', 'c', 'cpp', '--production-run']))
    args2 = config.resolve_config_parser(
        parser.parse_args(['--skip-languages', 'R', '--debug']))
    args3 = config.resolve_config_parser(
        parser.parse_args(['--validate-messages=True']))
    assert_equal(args3.validate_messages, True)
    old_var = {k: os.environ.get(k, None) for k in
               ['YGG_TEST_LANGUAGE', 'YGG_TEST_SKIP_LANGUAGE',
                'YGG_VALIDATE_COMPONENTS', 'YGG_VALIDATE_MESSAGES',
                'YGG_DEBUG', 'YGG_CLIENT_DEBUG']}
    args1.validate_messages = 'False'
    args2.validate_messages = 'True'
    with config.parser_config(args1):
        assert_equal(json.loads(os.environ.get('YGG_TEST_LANGUAGE', '')),
                     ['c', 'c++'])
        assert_equal(os.environ.get('YGG_VALIDATE_COMPONENTS', ''), 'false')
        assert_equal(os.environ.get('YGG_VALIDATE_MESSAGES', ''), 'false')
    for k, v in old_var.items():
        assert_equal(os.environ.get(k, None), v)
    with config.parser_config(args2):
        assert_equal(json.loads(os.environ.get('YGG_TEST_SKIP_LANGUAGE', '')),
                     ['R'])
        assert_equal(os.environ.get('YGG_DEBUG', ''), 'DEBUG')
        assert_equal(os.environ.get('YGG_CLIENT_DEBUG', ''), 'DEBUG')
        assert_equal(os.environ.get('YGG_VALIDATE_COMPONENTS', ''), 'true')
        assert_equal(os.environ.get('YGG_VALIDATE_MESSAGES', ''), 'true')
    for k, v in old_var.items():
        assert_equal(os.environ.get(k, None), v)
    assert_raises(ValueError, config.acquire_env,
                  dict(production_run=True, debug=True))
    with config.temp_config(production_run=True):
        assert_equal(os.environ.get('YGG_VALIDATE_COMPONENTS', ''), 'false')
        assert_equal(os.environ.get('YGG_VALIDATE_MESSAGES', ''), 'false')
    with config.temp_config(debug=True):
        assert_equal(os.environ.get('YGG_DEBUG', ''), 'DEBUG')
        assert_equal(os.environ.get('YGG_CLIENT_DEBUG', ''), 'DEBUG')
        assert_equal(os.environ.get('YGG_VALIDATE_COMPONENTS', ''), 'true')
        assert_equal(os.environ.get('YGG_VALIDATE_MESSAGES', ''), 'true')
    with config.temp_config(validate_messages='True'):
        assert_equal(os.environ.get('YGG_VALIDATE_MESSAGES', ''), 'true')



import uuid
import copy
from yggdrasil.communication.tests import test_CommBase as parent


class TestForkComm(parent.TestCommBase):
    r"""Tests for ForkComm communication class."""

    comm = 'ForkComm'
    attr_list = (copy.deepcopy(parent.TestCommBase.attr_list)
                 + ['comm_list', 'curr_comm_index'])
    ncomm = 2
    send_pattern = None
    recv_pattern = None
    test_error_send = None
    test_error_recv = None
    test_work_comm = None

    @property
    def cleanup_comm_classes(self):
        r"""list: Comm classes that should be cleaned up following the test."""
        return set([self.comm] + [None])

    @property
    def send_inst_kwargs(self):
        r"""dict: Keyword arguments for send instance."""
        out = super(TestForkComm, self).send_inst_kwargs
        out['ncomm'] = self.ncomm
        out['pattern'] = self.send_pattern
        return out

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for tested class."""
        out = super(TestForkComm, self).inst_kwargs
        out['pattern'] = self.recv_pattern
        return out

    def map_sent2recv(self, obj):
        r"""Convert a sent object into a received one."""
        if (((self.send_instance.pattern == 'scatter')
             and isinstance(obj, list) and obj)):
            single_obj = obj[0]
        else:
            single_obj = obj
        if self.recv_instance.pattern == 'gather':
            if obj in [b'', []]:
                return []
            return [single_obj for _ in range(self.ncomm)]
        return single_obj

    def duplicate_msg(self, out, direction='send'):
        r"""Copy a message for 'scatter' communication pattern."""
        if ((((direction == 'send')
              and (self.send_instance.pattern == 'scatter'))
             or ((direction == 'recv')
                 and (self.recv_instance.pattern == 'gather')))):
            out = [out for _ in range(self.ncomm)]
        return out
        
    @property
    def test_msg(self):
        r"""str: Test message that should be used for any send/recv tests."""
        return self.duplicate_msg(super(TestForkComm, self).test_msg)

    @property
    def test_msg_array(self):
        r"""str: Test message that should be used for any send/recv tests."""
        return self.duplicate_msg(super(TestForkComm, self).test_msg_array)

    @property
    def test_msg_dict(self):
        r"""dict: Test message that should be used for send_dict/recv_dict
        tests."""
        return self.duplicate_msg(super(TestForkComm, self).test_msg_dict)

    @property
    def msg_long(self):
        r"""str: Small test message for sending."""
        out = super(TestForkComm, self).test_msg
        if isinstance(out, bytes):
            out += (self.maxMsgSize * b'0')
        return self.duplicate_msg(out)
        
    @property
    def msg_filter_send(self):
        r"""object: Message to filter out on the send side."""
        return self.duplicate_msg(super(TestForkComm, self).msg_filter_send)
            
    @property
    def msg_filter_recv(self):
        r"""object: Message to filter out on the recv side."""
        return self.duplicate_msg(super(TestForkComm, self).msg_filter_recv)
        
    def test_error_name(self):
        r"""Test error on missing address."""
        self.assert_raises(RuntimeError, self.import_cls, 'test%s' % uuid.uuid4())

    def do_send_recv(self, *args, **kwargs):
        r"""Generic send/recv of a message."""
        if ((('eof' not in kwargs.get('send_meth', 'None'))
             and (not kwargs.get('no_recv', False))
             and (self.send_instance.pattern in ['broadcast', 'scatter'])
             and (self.recv_instance.pattern == 'cycle'))):
            kwargs.setdefault('n_recv', self.ncomm)
        super(TestForkComm, self).do_send_recv(*args, **kwargs)

    def test_send_recv_filter_eof(self, **kwargs):
        r"""Test send/recv of EOF with filter."""
        kwargs.setdefault('recv_timeout', 2 * self.timeout)
        super(TestForkComm, self).test_send_recv_filter_eof(**kwargs)
        
    def test_send_recv_filter_recv_filter(self, **kwargs):
        r"""Test send/recv with filter that blocks recv."""
        kwargs.setdefault('n_recv', 1)
        super(TestForkComm, self).test_send_recv_filter_recv_filter(**kwargs)
        
    def test_purge(self, **kwargs):
        r"""Test purging messages from the comm."""
        if self.send_instance.pattern == 'scatter':
            kwargs['msg_recv'] = [self.test_msg for _ in range(self.ncomm)]
        if (((self.send_instance.pattern in ['broadcast', 'scatter'])
             and (self.recv_instance.pattern == 'cycle'))):
            kwargs['nrecv'] = self.ncomm
        super(TestForkComm, self).test_purge(**kwargs)
        
    def test_send_recv_after_close(self, **kwargs):
        r"""Test that opening twice dosn't cause errors and that send/recv after
        close returns false."""
        kwargs.setdefault('msg_send', [self.test_msg for _ in range(self.ncomm)])
        super(TestForkComm, self).test_send_recv_after_close(**kwargs)


class TestForkCommList(TestForkComm):
    r"""Tests for ForkComm communication class with construction from address."""
    @property
    def inst_kwargs(self):
        r"""list: Keyword arguments for tested class."""
        out = super(TestForkComm, self).inst_kwargs
        out['comm_list'] = None  # To force test of construction from addresses
        return out


class TestForkCommCycle(TestForkComm):
    r"""Tests for ForkComm communication class with cycle/cycle communication
    pattern."""

    send_pattern = 'cycle'
    recv_pattern = 'cycle'


class TestForkCommScatter(TestForkComm):
    r"""Tests for ForkComm communication class with scatter/gather
    communication pattern."""
    
    send_pattern = 'scatter'
    recv_pattern = 'gather'

    def test_async_gather(self):
        r"""Test scatter-gather w/ intermittent send."""
        test_msg = self.test_msg
        self.send_instance.comm_list[0].send(test_msg[0])
        flag, msg_recv = self.recv_instance.recv()
        assert(flag)
        assert(self.recv_instance.is_empty_recv(msg_recv))
        for msg_send, comm in zip(test_msg[1:], self.send_instance.comm_list[1:]):
            assert(comm.send(msg_send))
        flag, msg_recv = self.recv_instance.recv()
        assert(flag)
        self.assert_msg_equal(msg_recv, test_msg)



import copy
from yggdrasil.communication import CommBase, get_comm, import_comm


_address_sep = ':YGG_ADD:'
_pattern_pairs = [('scatter', 'gather')]


class ForkedCommMessage(CommBase.CommMessage):
    r"""Class for forked comm messages.

    Args:
        msg (CommBase.CommMessage): Message being distributed.
        comm_list (list): List of communicators that the message is
            being distributed to.
        **kwargs: Additional keyword arguments are passed to the 'prepare_message'
            method for each communicator.

    """

    __slots__ = ['orig']

    def __init__(self, msg, comm_list, pattern='broadcast', **kwargs):
        super(ForkedCommMessage, self).__init__(
            msg=msg.msg, length=msg.length, flag=msg.flag,
            args=msg.args, header=msg.header)
        for k in CommBase.CommMessage.__slots__:
            setattr(self, k, getattr(msg, k))
        if (pattern in ['broadcast', 'cycle']) or (msg.flag == CommBase.FLAG_EOF):
            msg_list = [copy.deepcopy(msg)
                        for _ in range(len(comm_list))]
        elif pattern == 'scatter':
            msg_list = [copy.deepcopy(msg.args[i])
                        for i in range(len(comm_list))]
            kwargs.setdefault('flag', msg.flag)
            if msg.header:
                kwargs.setdefault('header_kwargs', msg.header)
        else:  # pragma: debug
            raise ValueError("Unsupported pattern: '%s'" % pattern)
        args = {i: x.prepare_message(msg_list[i], **kwargs)
                for i, x in enumerate(comm_list)}
        self.orig = msg.args
        self.args = args


def get_comm_name(name, i):
    r"""Get the name of the ith comm in the series.

    Args:
        name (str): Name of the fork comm.
        i (int): Index of comm in fork bundle.

    Returns:
        str: Name of ith comm in fork bundle.

    """
    return '%s_%d' % (name, i)


class ForkComm(CommBase.CommBase):
    r"""Class for receiving/sending messages from/to multiple comms.

    Args:
        name (str): The environment variable where communication address is
            stored.
        comm_list (list, optional): The list of options for the comms that
            should be bundled. If not provided, the bundle will be empty.
        pattern (str, optional): The communication pattern that should be
            used to handle outgoing/incoming messages. Options include:
              'cycle': Receive from or send to the next available comm in
                  the list (default for receiving comms).
              'broadcast': [SEND ONLY] Send the same message to each comm
                  (default for sending comms).
              'scatter': [SEND ONLY] Send part of message (must be a list)
                  to each comm.
              'gather': [RECV ONLY] Receive lists of messages from each
                  comm where a message is only returned when there is a
                  message from each.
        **kwargs: Additional keyword arguments are passed to the parent class.

    Attributes:
        comm_list (list): Comms included in this fork.
        curr_comm_index (int): Index comm that next receive will be from.

    """

    _commtype = 'fork'
    _dont_register = True
    child_keys = ['serializer_class', 'serializer_kwargs',
                  'format_str', 'field_names', 'field_units', 'as_array']
    noprop_keys = ['send_converter', 'recv_converter', 'filter', 'transform']
    
    def __init__(self, name, comm_list=None, is_async=False,
                 pattern=None, **kwargs):
        child_kwargs = {k: kwargs.pop(k) for k in self.child_keys if k in kwargs}
        noprop_kwargs = {k: kwargs.pop(k) for k in self.noprop_keys if k in kwargs}
        self.comm_list_backlog = {}
        self.comm_list = []
        self.curr_comm_index = 0
        self.eof_recv = []
        self.eof_send = []
        if pattern is None:
            if kwargs.get('direction', 'send') == 'recv':
                pattern = 'cycle'
            else:
                pattern = 'broadcast'
        self.pattern = pattern
        if kwargs.get('direction', 'send') == 'recv':
            assert(self.pattern in ['cycle', 'gather'])
        else:
            assert(self.pattern in ['cycle', 'scatter', 'broadcast'])
        address = kwargs.pop('address', None)
        if comm_list is None:
            if isinstance(address, list):
                ncomm = len(address)
            else:
                ncomm = 0
            comm_list = [None for i in range(ncomm)]
        assert(isinstance(comm_list, list))
        ncomm = len(comm_list)
        for i in range(ncomm):
            if comm_list[i] is None:
                comm_list[i] = {}
            if comm_list[i].get('name', None) is None:
                comm_list[i]['name'] = get_comm_name(name, i)
            for k in child_kwargs.keys():
                if k in comm_list[i]:  # pragma: debug
                    raise ValueError("The keyword '%s' was specified for both the "
                                     "root ForkComm and a child comm, but can only "
                                     "be present for one." % k)
        if isinstance(address, list):
            assert(len(address) == ncomm)
            for i in range(ncomm):
                comm_list[i]['address'] = address[i]
        for i in range(ncomm):
            ikw = copy.deepcopy(kwargs)
            ikw.update(child_kwargs)
            ikw.update(comm_list[i])
            ikw.setdefault('use_async', is_async)
            iname = ikw.pop('name')
            self.comm_list.append(get_comm(iname, **ikw))
            self.eof_recv.append(0)
            self.eof_send.append(0)
            self.comm_list_backlog[i] = []
        if ncomm > 0:
            kwargs['address'] = [x.address for x in self.comm_list]
        kwargs.update(noprop_kwargs)
        super(ForkComm, self).__init__(name, is_async=is_async, **kwargs)
        assert(not self.single_use)
        assert(not self.is_server)
        assert(not self.is_client)

    def disconnect(self):
        r"""Disconnect attributes that are aliases."""
        for x in self.comm_list:
            x.disconnect()
        super(ForkComm, self).disconnect()
        
    def printStatus(self, nindent=0, **kwargs):
        r"""Print status of the communicator."""
        super(ForkComm, self).printStatus(nindent=nindent,
                                          **kwargs)
        for x in self.comm_list:
            x.printStatus(nindent=nindent + 1)

    def __len__(self):
        return len(self.comm_list)

    @property
    def any_files(self):
        r"""bool: True if the comm interfaces with any files."""
        return any(x.is_file for x in self.comm_list)

    @property
    def last_comm(self):
        r"""CommBase: Last comm that was used."""
        idx = self.curr_comm_index
        if idx > 0:
            idx -= 1
        return self.comm_list[idx % len(self)]
    
    @property
    def curr_comm(self):
        r"""CommBase: Current comm."""
        return self.comm_list[self.curr_comm_index % len(self)]

    @property
    def maxMsgSize(self):
        r"""int: Maximum size of a single message that should be sent."""
        return min([x.maxMsgSize for x in self.comm_list])

    @classmethod
    def new_comm_kwargs(cls, name, *args, **kwargs):
        r"""Get keyword arguments for new comm."""
        if 'address' not in kwargs:
            addresses = []
            comm_list = kwargs.get('comm_list', None)
            ncomm = kwargs.pop('ncomm', 0)
            if comm_list is None:
                comm_list = [None for i in range(ncomm)]
            assert(isinstance(comm_list, list))
            ncomm = len(comm_list)
            for i in range(ncomm):
                x = comm_list[i]
                if x is None:
                    x = {}
                iname = x.pop('name', get_comm_name(name, i))
                icls = import_comm(x.get('commtype', None))
                _, ickw = icls.new_comm_kwargs(iname, **x)
                ickw['name'] = iname
                comm_list[i] = ickw
                addresses.append(ickw['address'])
            kwargs['comm_list'] = comm_list
            kwargs['address'] = addresses
        args = tuple([name] + list(args))
        return args, kwargs

    @property
    def model_env(self):
        r"""dict: Mapping between model name and opposite comm
        environment variables that need to be provided to the model."""
        out = {}
        for x in self.comm_list:
            iout = x.model_env
            for k, v in iout.items():
                out.setdefault(k, {})
                out[k].update(v)
        return out

    @property
    def opp_comms(self):
        r"""dict: Name/address pairs for opposite comms."""
        out = super(ForkComm, self).opp_comms
        out.pop(self.name)
        for x in self.comm_list:
            out.update(**x.opp_comms)
        return out

    def opp_comm_kwargs(self):
        r"""Get keyword arguments to initialize communication with opposite
        comm object.

        Returns:
            dict: Keyword arguments for opposite comm object.

        """
        kwargs = super(ForkComm, self).opp_comm_kwargs()
        kwargs['comm_list'] = [x.opp_comm_kwargs() for x in self.comm_list]
        for pair in _pattern_pairs:
            if self.pattern in pair:
                kwargs['pattern'] = pair[(pair.index(self.pattern) + 1) % 2]
        return kwargs

    def bind(self):
        r"""Bind in place of open."""
        for x in self.comm_list:
            x.bind()

    def open(self):
        r"""Open the connection."""
        for x in self.comm_list:
            x.open()

    def close(self, *args, **kwargs):
        r"""Close the connection."""
        for x in self.comm_list:
            x.close(*args, **kwargs)

    def close_in_thread(self, *args, **kwargs):  # pragma: debug
        r"""In a new thread, close the comm when it is empty."""
        raise Exception("ForkComm should not be closed in thread.")

    @property
    def is_open(self):
        r"""bool: True if the connection is open."""
        for x in self.comm_list:
            if x.is_open:
                return True
        return False

    @property
    def is_confirmed_send(self):
        r"""bool: True if all sent messages have been confirmed."""
        for x in self.comm_list:
            if not x.is_confirmed_send:  # pragma: debug
                return False
        return True

    @property
    def is_confirmed_recv(self):
        r"""bool: True if all received messages have been confirmed."""
        for x in self.comm_list:
            if not x.is_confirmed_recv:  # pragma: debug
                return False
        return True

    def confirm_send(self, noblock=False):
        r"""Confirm that sent message was received."""
        for x in self.comm_list:
            if not x.confirm_send(noblock=noblock):  # pragma: debug
                return False
        return True

    def confirm_recv(self, noblock=False):
        r"""Confirm that message was received."""
        for x in self.comm_list:
            if not x.confirm_recv(noblock=noblock):  # pragma: debug
                return False
        return True

    @property
    def n_msg_recv(self):
        r"""int: The number of incoming messages in the connection."""
        if self.pattern == 'gather':
            return min([x.n_msg_recv for x in self.comm_list])
        return sum([x.n_msg_recv for x in self.comm_list])

    @property
    def n_msg_send(self):
        r"""int: The number of outgoing messages in the connection."""
        if self.pattern in ['broadcast', 'scatter']:
            return max([x.n_msg_send for x in self.comm_list])
        return sum([x.n_msg_send for x in self.comm_list])

    @property
    def n_msg_recv_drain(self):
        r"""int: The number of incoming messages in the connection to drain."""
        return sum([x.n_msg_recv_drain for x in self.comm_list])

    @property
    def n_msg_send_drain(self):
        r"""int: The number of outgoing messages in the connection to drain."""
        return sum([x.n_msg_send_drain for x in self.comm_list])

    @property
    def empty_obj_recv(self):
        r"""obj: Empty message object."""
        if self.pattern in ['gather']:
            return []
        return self.last_comm.empty_obj_recv
        
    def update_serializer_from_message(self, msg):
        r"""Update the serializer based on information stored in a message.

        Args:
            msg (CommMessage): Outgoing message.

        """
        if msg.stype is not None:
            msg.stype = self.apply_transform_to_type(msg.stype)
        for x in self.comm_list:
            x.update_serializer_from_message(msg)
        
    def prepare_message(self, *args, **kwargs):
        r"""Perform actions preparing to send a message.

        Args:
            *args: Components of the outgoing message.
            **kwargs: Additional keyword arguments are passed to the prepare_message
                methods for the forked comms.

        Returns:
            CommMessage: Serialized and annotated message.

        """
        kws_root = {'skip_serialization': True}
        for k in ['header_kwargs']:
            if k in kwargs:
                kws_root[k] = kwargs.pop(k)
        msg = super(ForkComm, self).prepare_message(*args, **kws_root)
        if not isinstance(msg, ForkedCommMessage):
            msg = ForkedCommMessage(msg, self.comm_list,
                                    pattern=self.pattern, **kwargs)
        return msg

    def send_message(self, msg, **kwargs):
        r"""Send a message encapsulated in a CommMessage object.

        Args:
            msg (CommMessage): Message to be sent.
            **kwargs: Additional keyword arguments are passed to _safe_send.

        Returns:
            bool: Success or failure of send.
        
        """
        assert(isinstance(msg.args, dict))
        for idx in range(len(self)):
            i = self.curr_comm_index % len(self)
            x = self.curr_comm
            out = x.send_message(msg.args[i], **kwargs)
            self.errors += x.errors
            if msg.flag == CommBase.FLAG_EOF:
                self.eof_send[i] = 1
            self.curr_comm_index += 1
            if not out:
                return out
            elif (self.pattern == 'cycle') and (msg.flag != CommBase.FLAG_EOF):
                break
        msg.args = msg.orig
        msg.additional_messages = []
        kwargs['skip_safe_send'] = True
        return super(ForkComm, self).send_message(msg, **kwargs)
        
    def recv_message(self, *args, **kwargs):
        r"""Receive a message.

        Args:
            *args: Arguments are passed to the forked comm's recv_message method.
            **kwargs: Keyword arguments are passed to the forked comm's recv_message
                method.

        Returns:
            CommMessage: Received message.

        """
        timeout = kwargs.pop('timeout', None)
        if timeout is None:
            timeout = self.recv_timeout
        kwargs['timeout'] = 0
        first_comm = True
        T = self.start_timeout(timeout, key_suffix='recv:forkd')
        out = None
        out_gather = {}
        idx = None

        if self.pattern == 'gather':
            def complete():
                return (len(out_gather) == len(self))
        else:
            def complete():
                return bool(out_gather)
        
        while ((not T.is_out) or first_comm) and self.is_open and (not complete()):
            for i in range(len(self)):
                if complete():
                    break
                idx = self.curr_comm_index % len(self)
                x = self.curr_comm
                if idx not in out_gather:
                    if self.comm_list_backlog[idx]:
                        out_gather[idx] = self.comm_list_backlog[idx].pop(0)
                    elif x.is_open:
                        msg = x.recv_message(*args, **kwargs)
                        self.errors += x.errors
                        if msg.flag == CommBase.FLAG_EOF:
                            self.eof_recv[idx] = 1
                            if self.pattern == 'gather':
                                assert(all((v.flag == CommBase.FLAG_EOF)
                                           for v in out_gather.values()))
                                out_gather[idx] = msg
                            elif sum(self.eof_recv) == len(self):
                                out_gather[idx] = msg
                            else:
                                x.finalize_message(msg)
                        elif msg.flag == CommBase.FLAG_SUCCESS:
                            out_gather[idx] = msg
                self.curr_comm_index += 1
            first_comm = False
            if not complete():
                self.sleep()
        self.stop_timeout(key_suffix='recv:forkd')
        if complete():
            if self.pattern == 'cycle':
                idx, out = next(iter(out_gather.items()))
                out.args = {idx: copy.deepcopy(out)}
            elif self.pattern == 'gather':
                out = copy.deepcopy(next(iter(out_gather.values())))
                out.args = {idx: v for idx, v in out_gather.items()}
                # TODO: Gather header/type etc?
        else:
            for idx, v in out_gather.items():
                self.comm_list_backlog[idx].append(v)
            if self.is_closed:
                self.debug('Comm closed')
                out = CommBase.CommMessage(flag=CommBase.FLAG_FAILURE)
            else:
                out = CommBase.CommMessage(flag=CommBase.FLAG_EMPTY)
                if self.pattern == 'cycle':
                    out.args = self.last_comm.empty_obj_recv
                else:
                    out.args = []
        return out

    def finalize_message(self, msg, **kwargs):
        r"""Perform actions to decipher a message.

        Args:
            msg (CommMessage): Initial message object to be finalized.
            **kwargs: Keyword arguments are passed to the forked comm's
                finalize_message method.

        Returns:
            CommMessage: Deserialized and annotated message.

        """
        if msg.flag in [CommBase.FLAG_EOF, CommBase.FLAG_SUCCESS]:
            msg.args = {
                idx: self.comm_list[idx].finalize_message(
                    v, skip_python2language=True)
                for idx, v in msg.args.items()}
            if self.pattern == 'cycle':
                assert(len(msg.args) == 1)
                msg = next(iter(msg.args.values()))
            elif msg.flag == CommBase.FLAG_EOF:
                msg.args = msg.args[0].args
            else:
                msg.args = [msg.args[idx].args for idx in range(len(self))]
            msg.finalized = False
        return super(ForkComm, self).finalize_message(msg, **kwargs)
        
    @property
    def _multiple_first_send(self):  # pragma: debug
        return self.last_comm._multiple_first_send

    @_multiple_first_send.setter
    def _multiple_first_send(self, value):
        for x in self.comm_list:
            x._multiple_first_send = value

    @property
    def suppress_special_debug(self):
        return self.last_comm.suppress_special_debug

    @suppress_special_debug.setter
    def suppress_special_debug(self, value):
        for x in self.comm_list:
            x.suppress_special_debug = value

    @property
    def _type_errors(self):  # pragma: debug
        return self.last_comm._type_errors

    @_type_errors.setter
    def _type_errors(self, value):
        for x in self.comm_list:
            x._type_errors = value
    
    def purge(self):
        r"""Purge all messages from the comm."""
        super(ForkComm, self).purge()
        for x in self.comm_list:
            x.purge()
    
    def drain_server_signon_messages(self, **kwargs):
        r"""Drain server signon messages. This should only be used
        for testing purposes."""
        for x in self.comm_list:
            x.drain_server_signon_messages(**kwargs)

    def coerce_to_dict(self, msg, key_order, metadata):
        r"""Convert a message to a dictionary.

        Args:
            msg (object): Message to convert to a dictionary.
            key_order (list): Key order to use for the output dictionary.
            metadata (dict): Header data to accompany the message.

        Returns:
            dict: Converted message.

        """
        if self.pattern in ['scatter', 'gather']:
            assert(isinstance(msg, (list, tuple)) and (len(msg) == len(self)))
            out = [x.coerce_to_dict(msg[i], key_order, metadata)
                   for i, x in enumerate(self.comm_list)]
            return out
        return super(ForkComm, self).coerce_to_dict(msg, key_order, metadata)



import os
import copy
import pprint
import pystache
import yaml
import json
import git
import io as sio
from yggdrasil.schema import standardize, get_schema
from urllib.parse import urlparse
from yaml.constructor import (
    ConstructorError, BaseConstructor, Constructor, SafeConstructor)


class YAMLSpecificationError(RuntimeError):
    r"""Error raised when the yaml specification does not meet expectations."""
    pass


def no_duplicates_constructor(loader, node, deep=False):
    # https://gist.github.com/pypt/94d747fe5180851196eb
    """Check for duplicate keys."""
    mapping = {}
    for key_node, value_node in node.value:
        key = loader.construct_object(key_node, deep=deep)
        value = loader.construct_object(value_node, deep=deep)
        if key in mapping:
            raise ConstructorError("while constructing a mapping", node.start_mark,
                                   "found duplicate key (%s)" % key, key_node.start_mark)
        mapping[key] = value
    return loader.construct_mapping(node, deep)


for cls in (BaseConstructor, Constructor, SafeConstructor):
    cls.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                        no_duplicates_constructor)


def load_yaml(fname):
    r"""Parse a yaml file defining a run.

    Args:
        fname (str, file, dict): Path to a YAML file, an open file descriptor
            to a file containing a YAML, or a loaded YAML document. If fname starts with
            'git:' then the code will assume the file is in a remote git repository. The
            remainder of fname can be the full url to the YAML file
            (http://mygit.repo/foo/bar/yaml/interesting.yml) or just the repo and
            YAML file (the server is assumed to be github.com if not given)
            (foo/bar/yam/interesting.yaml will be interpreted as
            http://github.com/foo/bar/yam/interesting.yml).

    Returns:
        dict: Contents of yaml file.

    """
    opened = False
    if isinstance(fname, dict):
        yamlparsed = copy.deepcopy(fname)
        yamlparsed.setdefault('working_dir', os.getcwd())
        return yamlparsed
    elif isinstance(fname, str):
        # pull foreign file
        if fname.startswith('git:'):
            # drop the git prefix
            fname = fname[4:]
            # make sure we start with a full url
            if 'http' not in fname:
                url = 'http://github.com/' + fname
            else:
                url = fname
            # get the constituent url parts
            parsed = urlparse(url)
            # get the path component
            splitpath = parsed.path.split('/')
            # the first part is the 'owner' of the repo
            owner = splitpath[1]
            # the second part is the repo name
            reponame = splitpath[2]
            # the full path is the file name and location
            # turn the file path into an os based format
            fname = os.path.join(*splitpath)
            # check to see if the file already exists, and clone if it does not
            if not os.path.exists(fname):
                # create the url for cloning the repo
                cloneurl = parsed.scheme + '://' + parsed.netloc + '/' + owner + '/' +\
                    reponame
                # clone the repo into the appropriate directory
                _ = git.Repo.clone_from(cloneurl, os.path.join(owner, reponame))
                # now that it is cloned, just pass the yaml file (and path) onwards
        fname = os.path.realpath(fname)
        if not os.path.isfile(fname):
            raise IOError("Unable locate yaml file %s" % fname)
        fd = open(fname, 'r')
        opened = True
    else:
        fd = fname
        if (hasattr(fd, 'name') and (not fd.name.startswith('<'))):
            fname = fd.name
        else:
            fname = os.path.join(os.getcwd(), 'stream')
    # Mustache replace vars
    yamlparsed = fd.read()
    yamlparsed = pystache.render(
        sio.StringIO(yamlparsed).getvalue(), dict(os.environ))
    if fname.endswith('.json'):
        yamlparsed = json.loads(yamlparsed)
    else:
        yamlparsed = yaml.safe_load(yamlparsed)
    if not isinstance(yamlparsed, dict):  # pragma: debug
        raise YAMLSpecificationError("Loaded yaml is not a dictionary.")
    yamlparsed['working_dir'] = os.path.dirname(fname)
    if opened:
        fd.close()
    return yamlparsed


def prep_yaml(files):
    r"""Prepare yaml to be parsed by jsonschema including covering backwards
    compatible options.

    Args:
        files (str, list): Either the path to a single yaml file or a list of
            yaml files. Entries can also be opened file descriptors for files
            containing YAML documents or pre-loaded YAML documents.

    Returns:
        dict: YAML ready to be parsed using schema.

    """
    # Load each file
    if not isinstance(files, list):
        files = [files]
    yamls = [load_yaml(f) for f in files]
    # Load files pointed to
    for y in yamls:
        if 'include' in y:
            new_files = y.pop('include')
            if not isinstance(new_files, list):
                new_files = [new_files]
            for f in new_files:
                if not os.path.isabs(f):
                    f = os.path.join(y['working_dir'], f)
                yamls.append(load_yaml(f))
    # Standardize format of models and connections to be lists and
    # add working_dir to each
    comp_keys = ['models', 'connections']
    for yml in yamls:
        standardize(yml, comp_keys)
        for k in comp_keys:
            for x in yml[k]:
                if isinstance(x, dict):
                    x.setdefault('working_dir', yml['working_dir'])
    # Combine models & connections
    yml_all = {}
    for k in comp_keys:
        yml_all[k] = []
        for yml in yamls:
            yml_all[k] += yml[k]
    return yml_all


def parse_yaml(files, as_function=False):
    r"""Parse list of yaml files.

    Args:
        files (str, list): Either the path to a single yaml file or a list of
            yaml files.
        as_function (bool, optional): If True, the missing input/output channels
            will be created for using model(s) as a function. Defaults to False.

    Raises:
        ValueError: If the yml dictionary is missing a required keyword or has
            an invalid value.
        RuntimeError: If one of the I/O channels is not initialized with driver
            information.

    Returns:
        dict: Dictionary of information parsed from the yamls.

    """
    s = get_schema()
    # Parse files using schema
    yml_prep = prep_yaml(files)
    # print('prepped')
    # pprint.pprint(yml_prep)
    yml_norm = s.validate(yml_prep, normalize=True,
                          no_defaults=True, required_defaults=True)
    # print('normalized')
    # pprint.pprint(yml_norm)
    # Determine if any of the models require synchronization
    timesync_names = []
    for yml in yml_norm['models']:
        if yml.get('timesync', False):
            if yml['timesync'] is True:
                yml['timesync'] = 'timesync'
            if not isinstance(yml['timesync'], list):
                yml['timesync'] = [yml['timesync']]
            for i, tsync in enumerate(yml['timesync']):
                if isinstance(tsync, str):
                    tsync = {'name': tsync}
                    yml['timesync'][i] = tsync
                timesync_names.append(tsync['name'])
            yml.setdefault('timesync_client_of', [])
            yml['timesync_client_of'].append(tsync['name'])
    for tsync in set(timesync_names):
        for m in yml_norm['models']:
            if m['name'] == tsync:
                assert(m['language'] == 'timesync')
                m.update(is_server=True, inputs=[], outputs=[])
                break
        else:
            yml_norm['models'].append({'name': tsync,
                                       'args': [],
                                       'language': 'timesync',
                                       'is_server': True,
                                       'working_dir': os.getcwd(),
                                       'inputs': [],
                                       'outputs': []})
    # Parse models, then connections to ensure connections can be processed
    existing = None
    for k in ['models', 'connections']:
        for yml in yml_norm[k]:
            existing = parse_component(yml, k[:-1], existing=existing)
    # Add stand-in for function that will call to remote models
    if as_function:
        existing = add_model_function(existing)
    # Create server/client connections
    for srv, srv_info in existing['server'].items():
        clients = srv_info['clients']
        if srv not in existing['input']:
            continue
        yml = {'inputs': [{'name': x} for x in clients],
               'outputs': [{'name': srv}],
               'driver': 'RPCRequestDriver',
               'name': existing['input'][srv]['model_driver'][0]}
        if srv_info.get('replaces', None):
            yml['outputs'][0].update({
                k: v for k, v in srv_info['replaces']['input'].items()
                if k not in ['name']})
            yml['response_kwargs'] = {
                k: v for k, v in srv_info['replaces']['output'].items()
                if k not in ['name']}
        existing = parse_component(yml, 'connection', existing=existing)
        existing['model'][yml['dst_models'][0]]['clients'] = yml['src_models']
    existing.pop('server')
    # Make sure that servers have clients and clients have servers
    for k, v in existing['model'].items():
        if v.get('is_server', False):
            for x in existing['model'].values():
                if v['name'] in x.get('client_of', []):
                    break
            else:
                raise YAMLSpecificationError(
                    "Server '%s' does not have any clients.", k)
        elif v.get('client_of', False):
            for s in v['client_of']:
                missing_servers = []
                if s not in existing['model']:
                    missing_servers.append(s)
                if missing_servers:
                    print(list(existing['model'].keys()))
                    raise YAMLSpecificationError(
                        "Servers %s do not exist, but '%s' is a client of them."
                        % (missing_servers, v['name']))
    # Make sure that I/O channels initialized
    opp_map = {'input': 'output', 'output': 'input'}
    for io in ['input', 'output']:
        remove = []
        for k in list(existing[io].keys()):
            v = existing[io][k]
            if 'driver' not in v:
                if v.get('is_default', False):
                    remove.append(k)
                elif 'default_file' in v:
                    new_conn = {io + 's': [v['default_file']],
                                opp_map[io] + 's': [v]}
                    existing = parse_component(new_conn, 'connection',
                                               existing=existing)
                elif (io == 'input') and ('default_value' in v):
                    # TODO: The keys moved should be automated based on schema
                    # if the ValueComm has anymore parameters added
                    vdef = {'name': v['name'],
                            'default_value': v.pop('default_value'),
                            'count': v.pop('count', 1),
                            'commtype': 'value'}
                    new_conn = {'inputs': [vdef],
                                'outputs': [v]}
                    existing = parse_component(new_conn, 'connection',
                                               existing=existing)
                else:
                    raise YAMLSpecificationError(
                        "No driver established for %s channel %s" % (io, k))
        # Remove unused default channels
        for k in remove:
            for m in existing[io][k]['model_driver']:
                for i, x in enumerate(existing['model'][m][io + 's']):
                    if x['name'] == k:
                        existing['model'][m][io + 's'].pop(i)
                        break
            existing[io].pop(k)
    # Link io drivers back to models
    existing = link_model_io(existing)
    # print('drivers')
    # pprint.pprint(existing)
    return existing


def add_model_function(existing):
    r"""Patch input/output channels that are not connected to a function model.

    Args:
        existing (dict): Dictionary of existing components.

    Returns:
        dict: Updated dictionary of components.

    """
    new_model = {'name': 'function_model',
                 'language': 'function',
                 'args': 'function',
                 'working_dir': os.getcwd(),
                 'inputs': [],
                 'outputs': []}
    new_connections = []
    # Locate unmatched channels
    miss = {}
    dir2opp = {'input': 'output', 'output': 'input'}
    for io in dir2opp.keys():
        miss[io] = [k for k in existing[io].keys()
                    if not ((io == 'input') and (k in existing['server']))]
    for srv_info in existing['server'].values():
        if not srv_info['clients']:
            new_model.setdefault('client_of', [])
            new_model['client_of'].append(srv_info['model_name'])
    # for conn in existing['connection'].values():
    #     for io1, io2 in dir2opp.items():
    #         if ((io1 + 's') in conn):
    #             for x in conn[io1 + 's']:
    #                 if x in miss[io2]:
    #                     miss[io2].remove(x)
    # Create connections to function model
    for io1, io2 in dir2opp.items():
        for i in miss[io1]:
            function_channel = 'function_%s' % i
            function_comm = copy.deepcopy(existing[io1][i])
            function_comm['name'] = function_channel
            new_model[io2 + 's'].append(function_comm)
            new_connections.append({io1 + 's': [{'name': function_channel}],
                                    io2 + 's': [{'name': i}]})
    # Parse new components
    existing = parse_component(new_model, 'model', existing=existing)
    for new_conn in new_connections:
        existing = parse_component(new_conn, 'connection', existing=existing)
    return existing


def parse_component(yml, ctype, existing=None):
    r"""Parse a yaml entry for a component, adding it to the list of
    existing components.

    Args:
        yml (dict): YAML dictionary for a component.
        ctype (str): Component type. This can be 'input', 'output',
            'model', or 'connection'.
        existing (dict, optional): Dictionary of existing components.
            Defaults to empty dict.

    Raises:
        TypeError: If yml is not a dictionary.
        ValueError: If dtype is not 'input', 'output', 'model', or
            'connection'.
        ValueError: If the component already exists.

    Returns:
        dict: All components identified.

    """
    # s = get_schema()
    if not isinstance(yml, dict):
        raise YAMLSpecificationError("Component entry in yml must be a dictionary.")
    ctype_list = ['input', 'output', 'model', 'connection', 'server']
    if existing is None:
        existing = {k: {} for k in ctype_list}
    if ctype not in ctype_list:
        raise YAMLSpecificationError("'%s' is not a recognized component.")
    # Parse based on type
    if ctype == 'model':
        existing = parse_model(yml, existing)
    elif ctype == 'connection':
        existing = parse_connection(yml, existing)
    # elif ctype in ['input', 'output']:
    #     for k in ['inputs', 'outputs']:
    #         if k not in yml:
    #             continue
    #         for x in yml[k]:
    #             if 'commtype' not in x:
    #                 if 'filetype' in x:
    #                     x['commtype'] = s['file'].subtype2class[x['filetype']]
    #                 elif 'commtype' in x:
    #                     x['commtype'] = s['comm'].subtype2class[x['commtype']]
    # Ensure component dosn't already exist
    if yml['name'] in existing[ctype]:
        pprint.pprint(existing)
        pprint.pprint(yml)
        raise YAMLSpecificationError("%s is already a registered '%s' component." % (
            yml['name'], ctype))
    existing[ctype][yml['name']] = yml
    return existing


def parse_model(yml, existing):
    r"""Parse a yaml entry for a model.

    Args:
        yml (dict): YAML dictionary for a model.
        existing (dict): Dictionary of existing components.

    Returns:
        dict: Updated log of all entries.

    """
    _lang2driver = get_schema()['model'].subtype2class
    language = yml.pop('language')
    yml['driver'] = _lang2driver[language]
    # Add server input
    if yml.get('is_server', False):
        srv = {'name': '%s:%s' % (yml['name'], yml['name']),
               'commtype': 'default',
               'datatype': {'type': 'bytes'},
               'args': yml['name'] + '_SERVER',
               'working_dir': yml['working_dir']}
        if yml.get('function', False) and isinstance(yml['is_server'], bool):
            if (len(yml['inputs']) == 1) and (len(yml['outputs']) == 1):
                yml['is_server'] = {'input': yml['inputs'][0]['name'],
                                    'output': yml['outputs'][0]['name']}
            else:
                raise YAMLSpecificationError(
                    "The 'is_server' parameter is boolean for the model '%s' "
                    "and the 'function' parameter is also set. "
                    "If the 'function' and 'is_server' parameters are used "
                    "together, the 'is_server' parameter must be a mapping "
                    "with 'input' and 'output' entries specifying which of "
                    "the function's input/output variables should be received"
                    "/sent from/to clients. e.g. \n"
                    "\t-input: input_variable\n"
                    "\t-output: output_variables\n" % yml['name'])
        replaces = None
        if isinstance(yml['is_server'], dict):
            replaces = {}
            for io in ['input', 'output']:
                replaces[io] = None
                if not yml['is_server'][io].startswith('%s:' % yml['name']):
                    yml['is_server'][io] = '%s:%s' % (yml['name'], yml['is_server'][io])
                for i, x in enumerate(yml[io + 's']):
                    if x['name'] == yml['is_server'][io]:
                        replaces[io] = x
                        replaces[io + '_index'] = i
                        yml[io + 's'].pop(i)
                        break
                else:
                    raise YAMLSpecificationError(
                        "Failed to locate an existing %s channel "
                        "with the name %s." % (io, yml['is_server'][io]))
            srv['server_replaces'] = replaces
            yml['inputs'].insert(replaces['input_index'], srv)
        else:
            yml['inputs'].append(srv)
        yml['clients'] = []
        existing['server'].setdefault(srv['name'],
                                      {'clients': [],
                                       'model_name': yml['name']})
        if replaces:
            existing['server'][srv['name']]['replaces'] = replaces
    # Mark timesync clients
    timesync = yml.pop('timesync_client_of', [])
    if timesync:
        yml.setdefault('client_of', [])
        yml['client_of'] += timesync
    # Add client output
    if yml.get('client_of', []):
        for srv in yml['client_of']:
            srv_name = '%s:%s' % (srv, srv)
            if srv in timesync:
                cli_name = '%s:%s' % (yml['name'], srv)
            else:
                cli_name = '%s:%s_%s' % (yml['name'], srv, yml['name'])
            cli = {'name': cli_name,
                   'working_dir': yml['working_dir']}
            yml['outputs'].append(cli)
            existing['server'].setdefault(srv_name, {'clients': [],
                                                     'model_name': srv})
            existing['server'][srv_name]['clients'].append(cli_name)
    # Model index and I/O channels
    yml['model_index'] = len(existing['model'])
    for io in ['inputs', 'outputs']:
        for x in yml[io]:
            if yml.get('function', False) and (not x.get('outside_loop', False)):
                x.setdefault('dont_copy', True)
            if yml.get('allow_threading', False) or (
                    (yml.get('copies', 1) > 1)
                    and (not x.get('dont_copy', False))):
                x['allow_multiple_comms'] = True
            # TODO: Replace model_driver with partner_model?
            x['model_driver'] = [yml['name']]
            x['partner_model'] = yml['name']
            if yml.get('copies', 1) > 1:
                x['partner_copies'] = yml['copies']
            x['partner_language'] = language
            existing = parse_component(x, io[:-1], existing=existing)
    for k in yml.get('env', {}).keys():
        if not isinstance(yml['env'][k], str):
            yml['env'][k] = json.dumps(yml['env'][k])
    return existing


def parse_connection(yml, existing):
    r"""Parse a yaml entry for a connection between I/O channels.

    Args:
        yml (dict): YAML dictionary for a connection.
        existing (dict): Dictionary of existing components.

    Raises:
        RuntimeError: If the 'inputs' entry is not a model output or file.
        RuntimeError: If neither the 'inputs' or 'outputs' entries correspond
            to model I/O channels.

    Returns:
        dict: Updated log of all entries.

    """
    schema = get_schema()
    # File input
    is_file = {'inputs': [], 'outputs': []}
    iname_list = []
    for x in yml['inputs']:
        is_file['inputs'].append(schema.is_valid_component('file', x))
        if is_file['inputs'][-1]:
            fname = os.path.expanduser(x['name'])
            if not os.path.isabs(fname):
                fname = os.path.join(x['working_dir'], fname)
            fname = os.path.normpath(fname)
            if (not os.path.isfile(fname)) and (not x.get('wait_for_creation', False)):
                raise YAMLSpecificationError(
                    ("Input '%s' not found in any of the registered "
                     + "model outputs and is not a file.") % x['name'])
            x['address'] = fname
        elif 'default_value' in x:
            x['address'] = x['default_value']
        else:
            iname_list.append(x['name'])
    # File output
    oname_list = []
    for x in yml['outputs']:
        is_file['outputs'].append(schema.is_valid_component('file', x))
        if is_file['outputs'][-1]:
            fname = os.path.expanduser(x['name'])
            if not x.get('in_temp', False):
                if not os.path.isabs(fname):
                    fname = os.path.join(x['working_dir'], fname)
                fname = os.path.normpath(fname)
            x['address'] = fname
        else:
            oname_list.append(x['name'])
    iname = ','.join(iname_list)
    oname = ','.join(oname_list)
    if not iname:
        args = oname
    elif not oname:
        args = iname
    else:
        args = '%s_to_%s' % (iname, oname)
    name = args
    # Connection
    xx = {'src_models': [], 'dst_models': [],
          'inputs': [], 'outputs': []}
    for i, y in enumerate(yml['inputs']):
        if is_file['inputs'][i] or ('default_value' in y):
            xx['inputs'].append(y)
        else:
            new = existing['output'][y['name']]
            new.update(y)
            xx['inputs'].append(new)
            xx['src_models'] += existing['output'][y['name']]['model_driver']
            del existing['output'][y['name']]
    for i, y in enumerate(yml['outputs']):
        if is_file['outputs'][i]:
            xx['outputs'].append(y)
        else:
            new = existing['input'][y['name']]
            new.update(y)
            xx['outputs'].append(new)
            xx['dst_models'] += existing['input'][y['name']]['model_driver']
            del existing['input'][y['name']]
    # TODO: Split comms if models are not co-located and the main
    # process needs access to the message passed
    yml.update(xx)
    yml.setdefault('driver', 'ConnectionDriver')
    yml.setdefault('name', name)
    return existing


def link_model_io(existing):
    r"""Link I/O drivers back to the models they communicate with.

    Args:
        existing (dict): Dictionary of existing components.

    Returns:
        dict: Dictionary with I/O drivers added to models.

    """
    # Add fields
    for m in existing['model'].keys():
        existing['model'][m]['input_drivers'] = []
        existing['model'][m]['output_drivers'] = []
    # Add connections
    for io in existing['connection'].values():
        for m in io['src_models']:
            existing['model'][m]['output_drivers'].append(io)
        for m in io['dst_models']:
            existing['model'][m]['input_drivers'].append(io)
    return existing



import os
import tempfile
from yggdrasil.tests import assert_raises
from yggdrasil.metaschema.datatypes.tests import test_MetaschemaType as parent


class ValidClass(object):  # pragma: debug
    r"""Valid class for testing."""
    pass


class TestClassMetaschemaType(parent.TestMetaschemaType):
    r"""Test class for ClassMetaschemaType class."""

    _mod = 'ClassMetaschemaType'
    _cls = 'ClassMetaschemaType'

    @staticmethod
    def after_class_creation(cls):
        r"""Actions to be taken during class construction."""
        parent.TestMetaschemaType.after_class_creation(cls)
        cls._value = ValidClass
        cls._valid_encoded = [dict(cls._typedef,
                                   type=cls.get_import_cls().name)]
        cls._valid_decoded = [cls._value, object]
        cls._invalid_encoded = [{}]
        cls._invalid_decoded = ['hello']
        cls._compatible_objects = [(cls._value, cls._value, None)]
        cls._valid_normalize += [('%s:ValidClass' % __name__, ValidClass)]

    def test_module_file(self):
        r"""Test decoding data that includes the full path to the module file."""
        _tmpclassfile = os.path.join(tempfile.gettempdir(), 'test_class.py')
        with open(_tmpclassfile, 'w') as fd:
            fd.write('class ValidClass2(object):\n    pass\n')
        out = self.import_cls.decode_data('%s:ValidClass2' % _tmpclassfile, None)
        self.assert_equal(out.__name__, 'ValidClass2')

    def test_decode_data_errors(self):
        r"""Test errors in decode_data."""
        assert_raises(ValueError, self.import_cls.decode_data, 'hello', None)
        assert_raises(ImportError, self.import_cls.decode_data, 'invalid:invalid', None)
        assert_raises(AttributeError, self.import_cls.decode_data,
                      'yggdrasil:invalid', None)



from yggdrasil import tools
from yggdrasil.tests import assert_equal
from yggdrasil.communication import BufferComm
from yggdrasil.communication.tests import test_CommBase


def test_LockedBuffer_clear():
    r"""Test the clear method of the LockedBuffer class."""
    x = BufferComm.LockedBuffer()
    x.append('test')
    while x.empty():
        tools.sleep(0.1)
    assert(not x.empty())
    x.clear()
    assert_equal(len(x), 0)


def test_LockedBuffer_pop():
    r"""Test the pop method of the LockedBuffer class."""
    x = BufferComm.LockedBuffer()
    assert_equal(x.pop(default='hello'), 'hello')


class TestBufferComm(test_CommBase.TestCommBase):
    r"""Tests for BufferComm communication class."""

    comm = 'BufferComm'
    
    test_error_name = None



r"""This package provides a framework for integrating models across languages
such that they can be run simultaneously, passing input back and forth."""
import os
import sys
import shutil
from ._version import get_versions
from yggdrasil import platform, config
from yggdrasil.runner import YggFunction
_test_package_name = None
_test_package = None
config.cfg_logging()


if platform._is_win:  # pragma: windows
    # This is required to fix crash on Windows in case of Ctrl+C
    # https://github.com/ContinuumIO/anaconda-issues/issues/905#issuecomment-232498034
    os.environ['FOR_DISABLE_CONSOLE_CTRL_HANDLER'] = 'T'


if not os.path.isfile(config.usr_config_file):  # pragma: config
    from yggdrasil.languages import install_languages
    shutil.copy(config.def_config_file, config.usr_config_file)
    install_languages.install_all_languages(from_setup=True)
    if not any([x.endswith(('yggconfig', 'yggconfig.exe', 'config'))
                for x in sys.argv]):
        # Don't configure if that is what is going to happen anyway
        config.update_language_config(verbose=True)


def import_as_function(model_yaml, **kwargs):
    r"""Import a model as a function from a yaml specification file.

    Args:
        model_yaml (str, list): Full path to model yaml describing
            the model(s) and an inputs or outputs.
        **kwargs: Additional keyword arguments are passed to the
            YggFunction constructor.

    Returns:
        YggFunction: Callable wrapper for model.

    """
    return YggFunction(model_yaml, **kwargs)


__all__ = []
__version__ = get_versions()['version']
del get_versions



"""Constants used by yggdrasil."""


# ======================================================
# Do not edit this file past this point as the following
# is generated by yggdrasil.schema.update_constants
# ======================================================

LANG2EXT = {
    'R': '.R',
    'c': '.c',
    'c++': '.cpp',
    'cpp': '.cpp',
    'cxx': '.cpp',
    'executable': '.exe',
    'fortran': '.f90',
    'lpy': '.lpy',
    'matlab': '.m',
    'osr': '.xml',
    'python': '.py',
    'r': '.R',
    'sbml': '.xml',
    'yaml': '.yml',
}
EXT2LANG = {v: k for k, v in LANG2EXT.items()}
LANGUAGES = {
    'compiled': [
        'c', 'c++', 'fortran'],
    'interpreted': [
        'R', 'matlab', 'python'],
    'build': [
        'cmake', 'make'],
    'dsl': [
        'lpy', 'osr', 'sbml'],
    'other': [
        'executable', 'function', 'timesync'],
}
LANGUAGES['all'] = (
    LANGUAGES['compiled']
    + LANGUAGES['interpreted']
    + LANGUAGES['build']
    + LANGUAGES['dsl']
    + LANGUAGES['other'])
LANGUAGES_WITH_ALIASES = {
    'compiled': [
        'c', 'c++', 'cpp', 'cxx', 'fortran'],
    'interpreted': [
        'R', 'r', 'matlab', 'python'],
    'build': [
        'cmake', 'make'],
    'dsl': [
        'lpy', 'osr', 'sbml'],
    'other': [
        'executable', 'function', 'timesync'],
}
LANGUAGES_WITH_ALIASES['all'] = (
    LANGUAGES_WITH_ALIASES['compiled']
    + LANGUAGES_WITH_ALIASES['interpreted']
    + LANGUAGES_WITH_ALIASES['build']
    + LANGUAGES_WITH_ALIASES['dsl']
    + LANGUAGES_WITH_ALIASES['other'])
COMPILER_ENV_VARS = {
    'c': {
        'exec': 'CC',
        'flags': 'CFLAGS',
    },
    'c++': {
        'exec': 'CXX',
        'flags': 'CXXFLAGS',
    },
    'fortran': {
        'exec': 'FC',
        'flags': 'FFLAGS',
    },
}
COMPILATION_TOOL_VARS = {
    'LIB': {
        'exec': None,
        'flags': None,
    },
    'LINK': {
        'exec': 'LINK',
        'flags': 'LDFLAGS',
    },
    'ar': {
        'exec': 'AR',
        'flags': None,
    },
    'cl': {
        'exec': 'CC',
        'flags': 'CFLAGS',
    },
    'cl++': {
        'exec': 'CXX',
        'flags': 'CXXFLAGS',
    },
    'clang': {
        'exec': 'CC',
        'flags': 'CFLAGS',
    },
    'clang++': {
        'exec': 'CXX',
        'flags': 'CXXFLAGS',
    },
    'cmake': {
        'exec': None,
        'flags': None,
    },
    'g++': {
        'exec': 'CXX',
        'flags': 'CXXFLAGS',
    },
    'gcc': {
        'exec': 'CC',
        'flags': 'CFLAGS',
    },
    'gfortran': {
        'exec': 'FC',
        'flags': 'FFLAGS',
    },
    'ld': {
        'exec': 'LD',
        'flags': 'LDFLAGS',
    },
    'libtool': {
        'exec': 'LIBTOOL',
        'flags': None,
    },
    'make': {
        'exec': None,
        'flags': None,
    },
    'nmake': {
        'exec': None,
        'flags': None,
    },
}



import copy
import numpy as np
from yggdrasil.communication.transforms.TransformBase import TransformBase


class IterateTransform(TransformBase):
    r"""Class for iterating over message elements."""
    _transformtype = 'iterate'

    def validate_datatype(self, datatype):
        r"""Assert that the provided datatype is valid for this transformation.
        
        Args:
            datatype (dict): Datatype to validate.

        Raises:
            AssertionError: If the datatype is not valid.

        """
        assert(datatype.get('type', None) in ['array', 'object', '1darray',
                                              'ndarray'])

    def get_elements(self, datatype):
        r"""Get a list of elements in the datatype for iteration.

        Args:
            datatype (dict): Datatype to get elements from.

        Returns:
            list: List of datatypes for the elements iterated over.

        """
        if datatype.get('type', None) == 'array':
            if isinstance(datatype.get('items', None), dict):
                out = [datatype['items']]
            else:
                out = datatype['items']
        elif datatype.get('type', None) == 'object':
            out = [v for v in datatype.get('properties', {}).values()]
            if 'additionalProperties' in datatype:
                out.append(datatype['additionalProperties'])
        elif datatype.get('type', None) == '1darray':
            out = [dict(datatype, type='scalar')]
            out[0].pop('length', None)
        elif datatype.get('type', None) == 'ndarray':
            if len(datatype['shape']) > 2:
                out = [dict(datatype, shape=datatype['shape'][1:])]
            else:
                out = [dict(datatype, type='1darray',
                            length=datatype['shape'][-1])]
                out[0].pop('shape', None)
        else:  # pragma: debug
            raise ValueError("Unsupported datatype: %s" % datatype)
        return out
    
    def transform_datatype(self, datatype):
        r"""Determine the datatype that will result from applying the transform
        to the supplied datatype.

        Args:
            datatype (dict): Datatype to transform.

        Returns:
            dict: Transformed datatype.

        """
        elements = self.get_elements(datatype)
        if all(elements[0] == x for x in elements[1:]):
            datatype = copy.deepcopy(elements[0])
        else:
            datatype = {'type': 'any'}
        return datatype

    def evaluate_transform(self, x, no_copy=False):
        r"""Call transform on the provided message.

        Args:
            x (object): Message object to transform.
            no_copy (bool, optional): If True, the transformation occurs in
                place. Otherwise a copy is created and transformed. Defaults
                to False.

        Returns:
            object: The transformed message.

        """
        out = x
        if not no_copy:
            out = copy.deepcopy(out)
        if isinstance(out, dict):
            out = out.values()
        return iter(out)

    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the transform class.

        Returns:
            list: Multiple dictionaries of keywords and messages before/after
                pairs that will result from the transform created by the provided
                keywords.
        
        """
        return [
            {'kwargs': {
                'original_datatype': {
                    'type': 'array',
                    'items': {'type': 'int'}}},
             'in/out': [([0, 1, 2], iter([0, 1, 2]))],
             'in/out_t': [
                 ({'type': 'array', 'items': {'type': 'int'}},
                  {'type': 'int'})]},
            {'kwargs': {
                'original_datatype': {
                    'type': 'array',
                    'items': [{'type': 'int'}, {'type': 'int'}, {'type': 'int'}]}},
             'in/out': [([0, 1, 2], iter([0, 1, 2]))],
             'in/out_t': [
                 ({'type': 'array',
                   'items': [{'type': 'int'}, {'type': 'int'}, {'type': 'int'}]},
                  {'type': 'int'})]},
            {'kwargs': {
                'original_datatype': {
                    'type': 'array',
                    'items': [{'type': 'int'}, {'type': 'string'}]}},
             'in/out': [([0, 'hello'], iter([0, 'hello']))],
             'in/out_t': [
                 ({'type': 'array',
                   'items': [{'type': 'int'}, {'type': 'string'}]},
                  {'type': 'any'})]},
            {'kwargs': {
                'original_datatype': {
                    'type': 'object',
                    'properties': {'a': {'type': 'int'},
                                   'b': {'type': 'int'}},
                    'additionalProperties': {'type': 'int'}}},
             'in/out': [({'a': 0, 'b': 1, 'c': 2},
                         iter([0, 1, 2]))],
             'in/out_t': [
                 ({'type': 'object',
                   'properties': {'a': {'type': 'int'},
                                  'b': {'type': 'int'}},
                   'additionalProperties': {'type': 'int'}},
                  {'type': 'int'})]},
            {'kwargs': {
                'original_datatype': {
                    'type': '1darray', 'subtype': 'float',
                    'precision': 64, 'length': 3}},
             'in/out': [(np.ones(3, dtype='float64'),
                         iter(np.ones(3, dtype='float64')))],
             'in/out_t': [
                 ({'type': '1darray', 'subtype': 'float',
                   'precision': 64, 'length': 3},
                  {'type': 'scalar', 'subtype': 'float',
                   'precision': 64})]},
            {'kwargs': {
                'original_datatype': {
                    'type': 'ndarray', 'subtype': 'float',
                    'precision': 64, 'shape': [3, 4]}},
             'in/out': [(np.ones((3, 4), dtype='float64'),
                         iter(np.ones((3, 4), dtype='float64')))],
             'in/out_t': [
                 ({'type': 'ndarray', 'subtype': 'float',
                   'precision': 64, 'shape': [3, 4]},
                  {'type': '1darray', 'subtype': 'float',
                   'precision': 64, 'length': 4})]},
            {'kwargs': {
                'original_datatype': {
                    'type': 'ndarray', 'subtype': 'float',
                    'precision': 64, 'shape': [2, 3, 4]}},
             'in/out': [(np.ones((2, 3, 4), dtype='float64'),
                         iter(np.ones((2, 3, 4), dtype='float64')))],
             'in/out_t': [
                 ({'type': 'ndarray', 'subtype': 'float',
                   'precision': 64, 'shape': [2, 3, 4]},
                  {'type': 'ndarray', 'subtype': 'float',
                   'precision': 64, 'shape': [3, 4]})]},
        ]



import os
from yggdrasil.drivers.DSLModelDriver import DSLModelDriver


class SBMLModelDriver(DSLModelDriver):  # pragma: sbml
    r"""Class for running SBML models.

    Args:
        start_time (float, optional): Time that simulation should be
            started from. If 'reset' is True, the start time will
            always be the provided value, otherwise, the start time
            will be the end of the previous call after the first call.
            Defaults to 0.0.
        steps (int, optional): Number of steps that should be output.
            Defaults to None.
        reset (bool, optional): If True, the simulation will be reset
            to it's initial values before each call (including the
            start time). Defaults to False.
        selections (list, optional): Variables to include in the output.
            Defaults to None and the time/floating selections will be
            returned.
        integrator (str, optional): Name of integrator that should be
            used. Valid options include ['cvode', 'gillespie', 'rk4',
            'rk45']. Defaults to 'cvode'.
        integrator_settings (dict, optional): Settings for the
            integrator. Defaults to empty dict.
        variable_step (bool, optional): If True, the output steps will
            have a variable spacing. Defaults to False.
        only_output_final_step (bool, optional): If True, only the
            final timestep is output. Defaults to False.
        skip_start_time (bool, optional): If True, the results for the
            initial time step will not be output. Defaults to False.
            This option is ignored if only_output_final_step is True.

    """
    _schema_subtype_description = 'Model is an SBML model.'
    _schema_properties = {
        'start_time': {'type': 'number', 'default': 0.0},
        'steps': {'type': 'integer', 'default': 1},
        'reset': {'type': 'boolean', 'default': False},
        'selections': {'type': 'array', 'items': {'type': 'string'},
                       'default': []},
        'integrator': {'type': 'string', 'default': 'cvode',
                       'enum': ['cvode', 'gillespie', 'rk4', 'rk45']},
        'integrator_settings': {'type': 'object', 'default': {}},
        # 'variable_step': {'type': 'boolean', 'default': False},
        'skip_start_time': {'type': 'boolean', 'default': False},
        'only_output_final_step': {'type': 'boolean', 'default': False}}
    language = 'sbml'
    language_ext = '.xml'
    interface_dependencies = ['roadrunner']

    @classmethod
    def language_version(cls, **kwargs):
        r"""Determine the version of this language.

        Args:
            **kwargs: Keyword arguments are passed to cls.run_executable.

        Returns:
            str: Version of compiler/interpreter for this language.

        """
        try:
            import roadrunner
            return roadrunner.__version__
        except ImportError:  # pragma: debug
            raise RuntimeError("roadrunner not installed.")

    @property
    def model_wrapper_args(self):
        r"""tuple: Positional arguments for the model wrapper."""
        return (self.model_file, self.start_time, self.steps)

    @property
    def model_wrapper_kwargs(self):
        r"""dict: Keyword arguments for the model wrapper."""
        out = super(SBMLModelDriver, self).model_wrapper_kwargs
        out.update(
            {'inputs': self.inputs,
             'outputs': self.outputs,
             'integrator': self.integrator,
             'integrator_settings': self.integrator_settings,
             'working_dir': self.working_dir,
             'reset': self.reset, 'selections': self.selections,
             # 'variable_step': self.variable_step,
             'skip_start_time': self.skip_start_time,
             'only_output_final_step': self.only_output_final_step})
        return out
    
    @classmethod
    def model_wrapper(cls, model_file, start_time, steps,
                      inputs=[], outputs=[],
                      integrator=None, integrator_settings={},
                      env=None, working_dir=None, reset=False,
                      selections=None, variable_step=False,
                      skip_start_time=False,
                      only_output_final_step=False):
        r"""Model wrapper."""
        if env is not None:
            os.environ.update(env)
        if working_dir is not None:
            os.chdir(working_dir)
        curr_time = start_time
        model, input_map, output_map = cls.setup_model(
            model_file, inputs=inputs, outputs=outputs,
            integrator=integrator, integrator_settings=integrator_settings,
        )
        if not selections:
            selections = list(model.keys())
        if 'time' not in selections:
            selections = ['time'] + selections
        for k, v in output_map.items():
            if not v['vars']:
                v['vars'] = selections
        while True:
            time = None
            flag = False
            for k, v in input_map.items():
                flag, value = v['comm'].recv_dict(key_order=v['vars'])
                if not flag:
                    print("No more input from %s" % k)
                    break
                for iv in v['vars']:
                    if iv == 'time':
                        time = value[iv]
                    else:
                        model.setValue(iv, value[iv])
            if not flag:
                break
            if time is None:  # pragma: debug
                raise RuntimeError("No time variable recovered.")
            curr_time, out_value = cls.call_model(
                model, curr_time, time, steps,
                start_time=start_time,
                reset=reset, selections=selections,
                variable_step=variable_step)
            for k, v in output_map.items():
                if only_output_final_step:
                    iout = {iv: out_value[iv][-1] for iv in v['vars']}
                    nele = 1
                elif skip_start_time:
                    iout = {iv: out_value[iv][1:] for iv in v['vars']}
                    nele = steps
                else:
                    iout = {iv: out_value[iv] for iv in v['vars']}
                    nele = steps + 1
                if (nele > 1) and (not v['as_array']):
                    for i in range(nele):
                        iiout = {ik: iv[i] for ik, iv in iout.items()}
                        flag = v['comm'].send_dict(iiout, key_order=v['vars'])
                        if not flag:  # pragma: debug
                            raise RuntimeError("Error sending step %d to %s" % (i, k))
                else:
                    flag = v['comm'].send_dict(iout, key_order=v['vars'])
                    if not flag:  # pragma: debug
                        raise RuntimeError("Error sending to %s" % k)

    @classmethod
    def setup_model(cls, model_file, inputs=[], outputs=[],
                    integrator=None, integrator_settings={}):
        r"""Set up model class instance."""
        import roadrunner
        from yggdrasil.languages.Python.YggInterface import (
            YggInput, YggOutput)
        model = roadrunner.RoadRunner(model_file)
        if integrator is not None:
            model.setIntegrator(integrator)
        for k, v in integrator_settings.items():
            model.getIntegrator().setValue(k, v)
        input_map = {}
        output_map = {}
        for x in inputs:
            input_map[x['name']] = {
                'vars': x.get('vars', []),
                'comm': YggInput(x['name'], new_process=True)}
        for x in outputs:
            output_map[x['name']] = {
                'as_array': x.get('as_array', False),
                'vars': x.get('vars', []),
                'comm': YggOutput(x['name'], new_process=True)}
        return model, input_map, output_map

    @classmethod
    def call_model(cls, model, curr_time, end_time, steps,
                   reset=False, start_time=None, selections=None,
                   variable_step=False):
        r"""Call the model."""
        if reset:
            model.reset()
            curr_time = start_time
        out = model.simulate(float(curr_time), float(end_time),
                             selections=selections,
                             steps=int(steps))
        # Unsupported?
        # variableStep=variable_step)
        return end_time, out



import os
from yggdrasil.examples.tests.test_transforms import TestExampleTransforms


def dst(received_data):
    transform = os.environ['TEST_TRANSFORM']
    TestExampleTransforms.check_received_data(transform, received_data)
    print('CONFIRMED')
    return
